#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Méthodes utiles pour l'exploitation PyTorch

Created on Thu Oct 31 09:42:49 2019

@author: Cyrile Delestre
"""
from typing import Union, Optional, Tuple, List, Type, Dict, Any
from warnings import warn

import numpy as np
from pandas.core.frame import DataFrame

import torch

from dstk.utils import shape_list
from dstk.utils.deprecated import deprecated
from dstk.utils.errors import CheckError


def check_tensor(
        array: Union[torch.Tensor, np.ndarray, DataFrame, List, Tuple, int, float],
        dtype: Optional[Type]=None
    ):
    r"""
    Fonction vérifiant si le type de l'entrée est bien transformable en Tensor 
    PyTorch et, si c'est le cas, transformation de ce dernier en suivant le 
    type dtype. Si l'array n'est pas transformable, envoit d'une erreur de 
    type ErrorCheck.
    
    Parameters
    ----------
    array : Union[Tensor, np.ndarray, DataFrame, List, Tuple, int, float]
        checkage de l'array + conversion en Tensor.
    dtype : Optional[Type]
        cast vers un type PyTorch. Si None prend le type de l'array ou du 
        scalaire par défaut. Attention, type PyTorch (torch.int32, etc.)
    
    Returns
    -------
    tensor : Tensor
        array transformé en tensor au type dtype
    """
    if isinstance(array, torch.Tensor):
        return array if dtype is None else array.to(dtype)
    elif isinstance(array, np.ndarray):
        return (
            torch.from_numpy(array)
            if dtype is None
            else torch.from_numpy(array).to(dtype)
        )
    elif isinstance(array, DataFrame):
        return (
            torch.from_numpy(array.values)
            if dtype is None
            else torch.from_numpy(array.values).to(dtype)
        )
    elif isinstance(array, (list, tuple,)):
        return (
            torch.from_numpy(np.array(array))
            if dtype is None
            else torch.from_numpy(np.array(array)).to(dtype)
        )
    elif isinstance(array, (int, float, np.integer, np.floating,)):
        return (
            torch.Tensor(np.array([array]))
            if dtype is None
            else torch.Tensor(np.array([array])).to(dtype)
        )
    else:
        raise CheckError


def check_tensor_dict(data: Dict[str, Any], add_dim_batch: bool=False):
    r"""
    Fonction qui va transformer les éléments d'un dictionnaire en PyTorch 
    Tensor si c'est faisable.
    
    Parameters
    ----------
    data : Dict[str, Any]
        dictionnaire pouvant contenir ce que l'on souhaite.
    add_dim_batch : bool (=False)
        ajoute la dimension du batch si elle n'est pas présente.

    Returns
    -------
    data : Dict[str, Tensor]
        dictionnaire contenant des Tensor
    """
    def prepar_data(x):
        data_ = check_tensor(data[x])
        return data_.unsqueeze(0) if add_dim_batch else data_
    return {
        ii: prepar_data(ii)
        for ii in data 
        if isinstance(
            data[ii],
            (torch.Tensor, np.ndarray, list, tuple, DataFrame, int, float,
             np.integer, np.floating,)
        )
    }


@deprecated("Fonction plus utilisé, sera supprimée dans les MAJ à venir !")
def detach(array):
    r"""
    Fonction générique détachant les tensors du graph.
    
    Caution
    -------
    Fonction deprecated !
    
    Parameters
    ----------
    :array: liste/tuple de tensor, liste/tuple de liste/tuple de tensor,
        dictionnaire de tensor ou dictionnaire de liste/tuple de tensor.
    
    Returns
    -------
    Préservation de la structure de départ avec les tensors détachés du graph
    """
    if isinstance(array, (list, tuple)):
        if isinstance(array[0], torch.Tensor):
            res = []
            for aa in array:
                if aa.requires_grad:
                    res.append(aa.detach())
                else:
                    res.append(aa)
            if isinstance(array, tuple):
                return tuple(res)
            return res
        elif isinstance(array[0], (list, tuple)):
            if isinstance(array[0][0], torch.Tensor):
                res = []
                for aa in array:
                    res_int = []
                    for bb in aa:
                        if bb.requires_grad:
                            res_int.append(bb.detach())
                        else:
                            res_int.append(bb)
                    if isinstance(array[0], tuple):
                        res.append(tuple(res_int))
                    else:
                        res.append(res_int)
                if isinstance(array, tuple):
                    return tuple(res)
                return res
            else:
                raise TypeError(
                    "La fonction detach attend une liste/tuple de liste/tuple "
                    "de Tensor alors qu'il s'agit d'un type {} de {} de {}."
                    .format(
                        type(array),
                        type(array[0]),
                        type(array[0][0])
                    )
                )
        else:
            raise TypeError(
                "La fonction detach attend une liste/tuple de tensor ou une "
                "liste/tuple de liste/tuple alors qu'il s'agit d'un type {} "
                "de {}.".format(type(array), type(array))
            )
    elif isinstance(array, dict):
        items = list(array.items)
        if isinstance(items[0][0], (list, tuple)):
            if isinstance(items[0][0][0], torch.Tensor):
                res = []
                for key, aa in items:
                    res_int = []
                    for bb in aa:
                        if bb.requires_grad:
                            res_int.append(bb.detach())
                        else:
                            res_int.append(bb)
                    if isinstance(items[0][0], tuple):
                        res.append((key, tuple(res_int)))
                    else:
                        res.append((key, res_int))
                return dict(res)
            else:
                raise TypeError(
                    "La fonction detach attend un dictionnaire de liste/tuple "
                    "de Tensor alors qu'il s'agit d'un type {} de {} de {}."
                    .format(
                        type(array),
                        type(items[0][0]),
                        type(items[0][0][0])
                    )
                )
        elif isinstance(items[0][0], torch.Tensor):
            res = []
            for key, aa in items:
                if aa.requires_grad:
                    res.append((key, aa.detach()))
                else:
                    res.append((key, aa))
            return dict(res)
        else:
            raise TypeError(
                "La fonction detach attend un dictionnaire de Tensor alors "
                "qu'il s'agit d'un type {} de {}."
                .format(type(array), type(items[0][0]))
            )
    elif isinstance(array, torch.Tensor):
        if array.requires_grad:
            return array.detach()
        return array
    else:
        raise TypeError(
            "La fonction detach attend soit une liste/tuple, un dictionnaire "
            "ou un Tensor alors qu'il s'agit d'un type {}."
            .foramt(type(array))
        )


def collate_zeros_padding(
    sample: List[Dict[str, Union[torch.Tensor, np.ndarray]]],
    fields_series: List[Union[List, Tuple]],
    padding_value: float=0,
    torch_type: bool=True,
    batch_first: bool=True
):
    r"""
    Fonction de padding générique pour des séquences de différente longueur.
    
    Parameters
    ----------
    sample: List[Dict[str, Union[Tensor, np.ndarray]]]
        liste de dict des singletons composant le batch au format. Dans 
        un dictionnaire si size_field est défini utilise le size_field max
        de toute la list pour définir la taille max de l'array de la série. 
        Si  un array est plus grand il sera tronqué. Sinon prend le maximum 
        de  tous les arrays présente dans toutes les listes. Convertie ensuite 
        tous les champs en tensor. Exemple de sample :
            [{'a': np.array, 'b': np.array, 'siz_a': int, ...}, ...]
    fields_series : List[Union[List, Tuple]]
        liste de liste contenant le nom du champ de la série et 
        optionnellement le champ de la taille qui lui est associé. S'il n'y a 
        pas de champs associé à taille alors prend le maximum de la taille. 
        Par exemple avec l'exemple précédent :
            [['a', 'siz_a'], ['b']] ou  encore [['a', siz_a], ['b', None]]
    padding_value : float
        valeur des valeurs ajoutées. Par défaut 0.
    torch_type : bool
        retourne les vecteurs au format PyTorch (True) ou Numpy (False).
    batch_first : bool
        défini sur la première dimension est celle du batch (la seconde est 
        alors la série) ou l'inverse. Attention dans PyTorch par défaut dans 
        les RNN c'est l'inverse.
    
    Notes
    -----
    Il existe une fonction de zeros padding dans PyToch à surveiller. Mais 
    elle est pour le moment 10x plus lente que l'approche par completion d'un 
    array.
    
    Returns
    -------
    batch : Dict[str, Union[Tensor, np.ndarray]]
        dict au format :
            - {'a': tensor, 'b': tensor, 'siz_a': tensor, ...}
    """
    if len(sample) == 0:
        raise AttributeError(
            "L'ensemble des observations sample est vide : "
            f"len(sample) -> {len(sample)} !"
        )
    if not isinstance(sample[0], dict):
        raise AttributeError(
            "sample est une liste qui doit être composé de dictionnaire. Ce "
            f"n'est pas le cas : type(sample[0]) -> {type(sample[0])} !"
        )

    batch = {ff: None for ff in sample[0]}
    items_series = list(map(lambda x: x[0], fields_series))
    key_series = list(
        map(lambda x: x[1], filter(lambda y: len(y) == 2, fields_series))
    )

    for arg in batch:
        if arg in items_series:
            select_serie = list(
                filter(lambda x: x[0] == arg, fields_series)
            )[0]
            if len(select_serie) == 1 or select_serie[1] is None:
                key_len = None
                max_len = max(
                    sample,
                    key = lambda x: x[arg].shape[0]
                )[arg].shape[0]
            else:
                key_len = select_serie[1]
                max_len = max(
                    sample,
                    key = lambda x: x[key_len]
                )[key_len]

            obj_slice = None
            shape = sample[0][arg].shape
            if batch_first:
                global_shape = np.insert(shape, 0, len(sample))
                global_shape[1] = max_len
                if len(shape) == 1:
                    obj_slice = lambda ii, samp_len: np.s_[ii, :samp_len]
                elif len(shape) == 2:
                    obj_slice = lambda ii, samp_len: np.s_[ii, :samp_len, :]
            else:
                global_shape = np.insert(shape, 1, len(sample))
                global_shape[0] = max_len
                if len(shape) == 1:
                    obj_slice = lambda ii, samp_len: np.s_[:samp_len, ii]
                elif len(shape) == 2:
                    obj_slice = lambda ii, samp_len: np.s_[:samp_len, ii, :]
            if obj_slice is None:
                raise ValueError(
                    f"Dimension de série non supporté. La série {arg} est de "
                    "dimention len(sample[0][{arg}].shape) -> "
                    f"{len(sample[0][{arg}].shape)} !"
                )

            if isinstance(sample[0][arg], torch.Tensor):
                dtype = sample[0][arg].numpy().dtype
            else:
                dtype = sample[0][arg].dtype

            tensor = np.ones(global_shape, dtype=dtype)*padding_value

            for ii, samp in enumerate(sample):
                if key_len is None:
                    samp_len = samp[arg].shape[0]
                else:
                    samp_len = samp[key_len]
                tensor[obj_slice(ii, samp_len)] = samp[arg]
        else:
            if isinstance(sample[0][arg], (np.ndarray, torch.Tensor)):
                if isinstance(sample[0][arg], torch.Tensor):
                    dtype = sample[0][arg].numpy().dtype
                else:
                    dtype = sample[0][arg].dtype
                shape = np.insert(sample[0][arg].shape, 0, len(sample))
            elif isinstance(sample[0][arg], int):
                dtype = np.int32
                shape = (
                    (len(sample), 1,) if arg not in key_series
                    else (len(sample),)
                )
            elif isinstance(sample[0][arg], float):
                dtype = np.float32
                shape = (
                    (len(sample), 1,) if arg not in key_series
                    else (len(sample),)
                )
            else:
                raise ValueError(
                    f"Le type de l'argument {arg} est {type(sample[0][arg])} "
                    f"n'est pas pris en charge par l'implémentation. {arg} "
                    "doit être un ndarray, Tensor, int ou float."
                )

            if len(shape) == 1:
                obj_slice = lambda ii: np.s_[ii]
            elif len(shape) == 2:
                obj_slice = lambda ii: np.s_[ii, :]
            else:
                raise ValueError(
                    f"Dimension de série non supporté. La série {arg} est de "
                    f"dimention len(sample[0][{arg}].shape) -> "
                    f"{len(sample[0][arg].shape)} !"
                )

            tensor = np.zeros(shape, dtype=dtype)
            for ii, samp in enumerate(sample):
                tensor[obj_slice(ii)] = samp[arg]

        batch[arg] = torch.from_numpy(tensor) if torch_type else tensor
    return batch


def make_mask(
        in_features: int,
        out_features: int,
        n_channel: int,
        shift: int,
        kernel_size: int,
        repeat: Optional[Union[List, Tuple]]=None
) -> torch.BoolTensor:
    r"""
    Fonction de création de masque permettant de personaliser la topologie 
    des réseaux linéaires pour être autoregressif.
    
    Parameters
    ----------
    in_features : int
        taille en entré du réseau
    out_features : int
        taille en sortie du réseau
    n_channel : int
        nombre de neuronnes voyant les mêmes entrées
    shift : int
        décalage incrémentale sur chaques neuronnes
    kernel_size : int
        entrée contigue en entré du neuronne
    repeat : Optional[Union[List, Tuple]] (=None)
        tuple ou liste permettant de constituer les connexions (utile si 
        entrés multivariées).
    
    Returns
    -------
    mask : BoolTensor
        masque de dimension (out_feature x repeat[0]), 
        (in_features x repeat[1])
    
    Examples
    --------
    >>> import torch
    >>> from dstk.pytorch import make_mask
    >>> 
    >>> make_mask(6,6,1,0,2, repeat=(1,2)).type(torch.int8)
    tensor([[1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0]], dtype=torch.int8)
    >>> 
    >>> make_mask(6,6,2,2,2, repeat=(1,2)).type(torch.int8)
    tensor([[1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0],
            [0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1],
            [0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1]], dtype=torch.int8)
    """
    mask = torch.zeros(out_features, in_features, dtype=torch.bool)
    for ii in range(0, out_features, n_channel):
        ii_shift = int(ii/n_channel)*shift
        mask[ii:ii+n_channel, ii_shift:ii_shift+kernel_size] = True
    if repeat:
        return mask.repeat(repeat)
    return mask



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Modèles de réseaux MLP génériques utilisables pour PyTorch

Created on Mon Sep 16 14:42:57 2019

@author: Cyrile Delestre
"""
from typing import Optional, Callable, Union, Dict, Any, List, TypeVar, Tuple

import numpy as np

import torch
from torch.nn import (Module, Linear, LeakyReLU, ReLU, Dropout, Identity,
                      BatchNorm1d, LayerNorm, ModuleList, functional as F)

from dstk.pytorch import make_mask

TMLP = TypeVar('TMLP', bound='MLP')
TRNLL = TypeVar('TRNLL', bound='RisudualNonLinearLayer')

class MLP(Module):
    r"""
    Classe MLP de construction automatique en fonction du nombre de hidden
    layers et de la topologie des dimensions.
    
    Parameters
    ----------
    dim_in : int
        dimension d'entrée
    dim_out : int
        dimension de sortie
    dim_first_lay : int
        nombre de neronne de la première couche
    n_layers : int
        nombre de hidden layers
    embed_topo : str
        topologie des dimensions des hidden layers
            - linear: part de in_units et progresse linéairement vers 
            out_unist
            - bottleneck: part de in_units et progresse linéairement vers 
                inter_units puis progresse linéaiement vers out_units
        (par défaut 'linear')
    inter_units : int
        dimension intermédiaire pour l'option
        embed_topo = 'bottleneck', sinon inutile (par défaut 10)
    alpha : float
        coefficient de non-linéarité de la fonction d'activation LeakyReLU
        (par défaut 0.3)
    dropout_prob : float
        proportion de dropout en proba (par défaut 0 <= pas de dropout)
    batchnorm : bool
        utilise la batchnorm (par défaut True)
    dropout : bool
        utilise le dropout (par défaut True)
    batchnorm_last_layer : bool
        applique la batchnormalization sur la dernière couche (par défaut 
        True)
    activation_last_layer : Optional[Callable]
        permet d'envoyer une fonction d'activation sur la dernière couche 
        (logit). Si None, pas de fonction d'activation appliquée.
    dropout_last_layer : bool
        applique ou non le dropout sur la dernière couche (par défaut True)
    
    Returns
    -------
    res : Tensor
        La dimension de sortie est (None, out_units)
    """
    def __init__(
        self,
        dim_in: int,
        dim_out: int,
        dim_first_lay: int,
        n_layers: int,
        embed_topo: str='linear',
        inter_units: int=10,
        alpha: float=0.3,
        dropout_prob: float=0,
        batchnorm: bool=True,
        dropout: bool=True,
        batchnorm_last_layer: bool=True,
        activation_last_layer: Optional[Callable]=None,
        dropout_last_layer: bool=True
    ):
        assert embed_topo in ['linear', 'bottleneck'] \
            and dim_in > 0 \
            and dim_out > 0
        super(MLP, self).__init__()
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.dim_first_lay = dim_first_lay
        self.n_layers = n_layers
        self.embed_topo = embed_topo
        self.inter_units = inter_units
        self.alpha = alpha
        self.dropout_prob = dropout_prob
        self.batchnorm = batchnorm
        self.dropout = dropout
        self.batchnorm_last_layer = batchnorm_last_layer
        self.activation_last_layer = activation_last_layer
        self.dropout_last_layer = dropout_last_layer
        self.build()

    def build(self) -> TMLP:
        r"""
        Fonction de construction du réseau de neurones.
        
        Warnings
        --------
        Pour Scikit-Learn, il faut que la fonction d'initialisation soit 
        différente de la fonction de construction.
        """
        if isinstance(self.activation_last_layer, bool):
            raise ValueError(
                "activation_last_layer est un booléen. Ce paramètre doit "
                "être une fonction ou None si aucune activation_last_layer "
                "ne doit être appliquée."
            )

        if self.n_layers > 1:
            if self.embed_topo == 'linear':
                    d_unite = (
                        (self.dim_first_lay-self.dim_out)/(self.n_layers-1)
                    )
                    units_by_lay = [
                        int(self.dim_first_lay-ii*d_unite)
                        for ii in range(self.n_layers)
                    ] 
            elif self.embed_topo == 'bottleneck':
                d_unite_1 = (
                    (self.dim_first_lay-self.inter_units)/
                    np.floor(self.n_layers/2)
                )
                units_dec = [
                    int(self.dim_first_lay-ii*d_unite_1)
                    for ii in range(int(np.floor(self.n_layers/2)))
                ]
                if self.n_layers == 2:
                    units_gro = [self.dim_out]
                else:
                    d_unite_2 = (
                        (self.inter_units-self.dim_out)/
                        np.ceil(self.n_layers/2-1)
                    )
                    units_gro = [
                        int(self.inter_units-ii*d_unite_2)
                        for ii in range(int(np.ceil(self.n_layers/2)))
                    ]
                units_by_lay = units_dec+units_gro
            else:
                raise AttributeError(
                    "Erreur embed_topo not in ['linear', 'bottleneck']"
                )
            units_in = [self.dim_in]+units_by_lay[:-1]
            units_out = units_by_lay
        else:
            units_in = [self.dim_in]
            units_out = [self.dim_out]
        layers = []
        for ii, nn in enumerate(zip(units_in, units_out)):
            n_in, n_out = nn
            layers.append(
                Linear(
                    in_features = n_in,
                    out_features = n_out,
                    bias = (
                        not self.batchnorm 
                        or (ii == self.n_layers-1 
                            and not self.batchnorm_last_layer)
                    )
                )
            )
            if (
                (ii < self.n_layers-1 or self.batchnorm_last_layer)
                and self.batchnorm
            ):
                layers.append(BatchNorm1d(num_features=n_out))
            if ii < self.n_layers-1:
                layers.append(LeakyReLU(negative_slope=self.alpha))
            if (
                ii == self.n_layers-1 and
                self.activation_last_layer is not None
            ):
                layers.append(self.activation_last_layer)
            if (
                (ii < self.n_layers-1 or self.dropout_last_layer)
                and self.dropout
            ):
                layers.append(Dropout(p=self.dropout_prob))
        self.layers = ModuleList(layers)
        return self

    def forward(self, inputs: torch.Tensor, **kargs) -> torch.Tensor:
        r"""
        Fonction d'appel du MLP
        
        Parameters
        ----------
        inputs : Tensor
            inputs de dimensions (batch_size, feat_size)
        
        Returns
        -------
        x_lay : Tensor
            sortie du MLP
        """
        x_lay = inputs
        for lay in self.layers:
            x_lay = lay(x_lay)
        return x_lay


class MaskedLinear(Linear):
    r"""
    Surcharge de la couche linéaire de PyTorch permettant de réaliser un 
    masque sur les poids afin de personaliser les connexions des neuronnes 
    composant la couche.
    Typiquement l'opération appliqué sera :
        input @ (mask*weight).T + bias
    Pour plus d'information sur la couche linéaire Linear de PyTorch, ce 
    référer à `la doc Linear`_.
    
    .. la doc Linear: https://pytorch.org/docs/stable/generated/torch.nn.Linear.html#torch.nn.Linear
    
    Parameters
    ----------
    in_features : int
        taille de l'entée de la couche linénaire
    out_features : int
        taille de la sortie de la couche linéaire
    bias : bool (=True)
        application d'un biais à la couche (True) ou non (False)
    """
    def __init__(self, in_features: int, out_features: int, bias: bool=True):
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            bias=bias
        )
        self.register_buffer(
            'mask',
            torch.ones(out_features, in_features, dtype=torch.bool)
        )

    def set_mask(self, mask: Union[torch.BoolTensor, np.ndarray]) -> None:
        r"""
        Application d'un masque afin de personnaliser les connexions.
        
        Parameters
        ----------
        mask : Union[torch.BoolTensor, np.ndarray]
            Masque (out_features x in_features) qui doit être un numpy array 
            ou un tensor Booléen. 
        """
        if isinstance(mask, np.ndarray):
            mask = torch.from_numpy(mask.astype(np.bool))
        self.mask.data.copy_(mask)
        return self

    def forward(self, input: torch.Tensor, **kargs) -> torch.Tensor:
        return F.linear(input, self.mask*self.weight, self.bias)


class MaskedFeedForward(Module):
    r"""
    Couche non liénaire masqué à 1 couche cachée, inspiré de la couche 
    FeedForward de la modélisation détection de multi-attention Transformer.
    
    Paramreters
    -----------
    in_features : int
        taille de l'entée de la couche
    latent_features : int
        taille de l'espace latent
    in_kernel_size : int
        entrée contigue en entré du neuronne
    latent_channel : int
        nombre de neuronnes voyant les mêmes entrées de l'espace latent
    shift : int
        décalage incrémentale sur chaques neuronnes
    repeat: Optional[Union[List, Tuple]]
        répétition du masque si modèle multié série
    dropout_prob : float (=0.3)
        probabilité de dropout des différentes étages du réseau
    activation : Callable (=LeakyReLU(0.3))
        couche non linéaire d'activation
    """
    def __init__(
        self,
        in_features: int,
        latent_features: int,
        in_kernel_size: int,
        latent_channel: int,
        shift: int,
        repeat: Optional[Union[List, Tuple]]=None,
        dropout_prob: float=0.3,
        activation: Callable=LeakyReLU(0.3)
    ):
        super().__init__()
        self.in_features = in_features
        self.latent_features = latent_features
        self.in_kernel_size = in_kernel_size
        self.latent_channel = latent_channel
        self.shift = shift
        self.repeat = repeat
        self.dropout_prob = dropout_prob
        self.activation = activation
        self.build()

    def build(self) -> TRNLL:
        r"""
        Fonction de construction du réseau de neurones.
        
        Warnings
        --------
        Pour Scikit-Learn, il faut que la fonction d'initialisation soit 
        différente de la fonction de construction.
        """
        mask = make_mask(
            in_features=self.in_features,
            out_features=self.latent_features,
            n_channel=self.num_channel,
            shift=self.shift,
            repeat=self.repeat
        )
        self.linear_in = MaskedLinear(
            in_features=self.in_features,
            out_features=self.latent_features
        ).set_mask(mask)
        self.linear_out = MaskedLinear(
            in_features=self.latent_features,
            out_features=self.out_features
        ).set_mask(mask.T)
        self.norm_in = LayerNorm(self.in_features)
        self.norm_out = LayerNorm(self.out_features)
        self.dropout1 = Dropout(self.dropout_prob)
        self.dropout2 = Dropout(self.dropout_prob)
        return self

    def forward(self, input: torch.Tensor, **kargs) -> torch.Tensor:
        r"""
        Fonction d'appel du MaskedFeedForward
        
        Parameters
        ----------
        inputs : Tensor
            inputs de dimensions (batch_size, feat_size)
        
        Returns
        -------
        x : Tensor
            sortie du RisudualNonLinearLayer
        """
        input = self.norm_in(input)
        x = self.linear_out(
            self.dropout1(
                self.activation(
                    self.linear_in(input)
                )
            )
        )
        return self.norm_out(self.dropout2(x+input))

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features}, "
            f"latent_features={self.latent_features}, "
            f"in_kernel_size={self.in_kernel_size}, "
            f"latent_channel={self.latent_channel}, "
            f"shift={self.shift}, "
            f"dropout_prob={self.dropout_prob}, "
            f"activation={self.activation}"
        )


class MaskedResidualLayer(Module):
    r"""
    Couche rédiuelle inspérée par le LSTM et utilisé notamment dans PixelNet 
    et WaveNet.
    
    Parameters
    ----------
    in_features : int
        taille de l'entée de la couche
    out_features : int
        taille de l'espace latent
    kernel_size : int
        entrée contigue en entré du neuronne
    shift : int
        décalage incrémentale sur chaques neuronnes
    repeat : Optional[Union[List, Tuple]]
        répétition du masque si modèle multié série
    dropout_prob : float (=0.3)
        probabilité de dropout des différentes étages du réseau
    """
    def __init__(self,
        in_features: int,
        out_features: int,
        kernel_size: int,
        shift: int,
        repeat: Optional[Union[List, Tuple]]=None,
        dropout_prob: float=0.3
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.kernel_size = kernel_size
        self.shift = shift
        self.repeat = repeat
        self.dropout_prob = dropout_prob
        self.build()
    
    def build(self):
        r"""
        Fonction de construction du réseau de neurones.
        
        Warnings
        --------
        Pour Scikit-Learn, il faut que la fonction d'initialisation soit 
        différente de la fonction de construction.
        """
        mask = make_mask(
            in_features=self.in_features,
            out_features=self.out_features,
            n_channel=1,
            shift=self.shift,
            repeat=self.repeat
        )
        self.linear1 = MaskedLinear(
            in_features=self.in_features,
            out_features=self.latent_features
        ).set_mask(mask)
        self.linear2 = MaskedLinear(
            in_features=self.in_features,
            out_features=self.latent_features
        ).set_mask(mask)
        self.norm_in = LayerNorm(self.in_features)
        self.norm_out = LayerNorm(self.out_features)
        self.dropout1 = Dropout(self.dropout_prob)
        self.dropout2 = Dropout(self.dropout_prob)
        return self
    
    def forward(self, input: torch.Tensor, **kargs) -> torch.Tensor:
        r"""
        Fonction d'appel du MaskedResidualLayer
        
        Parameters
        ----------
        inputs : Tensor
            inputs de dimensions (batch_size, feat_size)
        
        Returns
        -------
        x : Tensor
            sortie du RisudualNonLinearLayer
        """
        input = self.norm_in(input)
        x = (
            self.dropout1(F.sigmoid(self.linear1(input))) 
            * self.dropout2(F.tanh(self.linear2(input)))
        )
        return self.norm_out(x + input)


class Autoregressive(Module):
    def __init__(
        self,
        dim_in: int,
        dim_out: int,
        connexion_layer: List[Dict[str, Any]]=[
            dict(n_layers=1, n_channel=1, shift=5, kernel_size=5),
            dict(n_layers=1, n_channel=1, shift=2, kernel_size=2),
            dict(n_layers=1, n_channel=1, shift=2, kernel_size=2)
        ],
        module_layer: Module=MaskedResidualLayer,
        last_activation: Module=Identity
    ):
        super().__init__()
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.connexion_layer = connexion_layer
        self.module_layer = module_layer
        self.last_activation = last_activation
        self.build()
    
    def build(self):
        ...

    def forward(self, input: torch.Tensor, **kargs) -> torch.Tensor:
        ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Fonctions de checkage
^^^^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 11:55:58 2020

@author: Cyrile Delestre
"""

from typing import Any, Union, Optional, List, Tuple

import pandas as pd

from dstk.utils.errors import NoConsistentError

def check_dataframe(X: Union[pd.DataFrame, Any],
                    columns_name: Optional[Union[str, list]]=None,
                    copy=False):
    """
    Foncton permettant de checker si la variable X est bien une DataFrame et
    optionnellement vérifie si la colonne ou la liste des colonnes de 
    columns_name est présente dans la DataFrame ou non. Cette fonction 
    permet de sortir également un copy de la DataFrame.
    
    Parameters
    ----------
    X : Union[pd.DataFrame, Any]
        variable à tester.
    columns_name : Optional[Union[str, list]]
        noms des colonnes à tester la présence dans a DataFrame X.
    copy: bool
        si True sort une copie de X.
    
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dstk.utils import check_dataframe
    
    >>> array = np.array([1, 2, 3])
    >>> check_dataframe(array)
    AttributeError: L'argument d'entrée n'est pas une DataFrame mais une 
    <class 'numpy.ndarray'>.
    
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': ['d', 'e', 'f']})
    >>> check_dataframe(df, ['a', 'c'])
    AttributeError: Les colonnes c ne sont pas présentes dans la DataFrame qui 
    a pour colonnes : a, b.
    """
    if not isinstance(X, pd.DataFrame):
        raise AttributeError(
            f"L'argument d'entrée n'est pas une DataFrame mais une {type(X)}."
        )
    if isinstance(columns_name, (list, tuple)):
        check = list(filter(lambda x: not x in X.columns, columns_name))
        if len(check) > 0:
            raise AttributeError(
                f"Les colonnes {', '.join(check)} ne sont pas présentes dans "
                "la DataFrame qui a pour colonnes : "
                f"{', '.join(X.columns.tolist())}."
            )
    elif isinstance(columns_name, str):
        if not columns_name in X.columns:
            raise AttributeError(
                f"La colonne {columns_name} n'est pas présente dans la "
                "DataFrame qui possède comme colonnes : "
                f"{', '.join(X.columns.tolist())}."
            )
    elif columns_name is not None:
        raise AttributeError(
            "columns_name doit être un string, un tuple ou une liste et non "
            f"de type {type(columns_name)}."
        )
    if copy:
        return X.copy()

def shape_list(l: Union[List[Any], Tuple[Any]]):
    r"""
    Fonction récurcive permettant de déterminer la dimension d'une liste ou 
    d'un tuple. Fonction retournant un résultat équivalent à shape du Numpy.
    
    Parameters
    ----------
    l : Union[List[Any], Tuple[Any]]
        liste ou typle de dimension inconue
    
    Returns
    -------
    shape : List[int]
        list contenant les dimensions de la liste ou du tuple.
    
    Examples
    --------
    >>> from dstk.utils import shape_list
    
    >>> my_list
    [[[3, 7], [1, 2], [2, 3]], [[7, 2], [3, 1], [0, 7]]]
    >>> shape_list(my_list)
    [2, 3, 2]
    
    >>> my_list_bis
    [[[3, 7], [1, 2], [2]], [[7, 2], [3, 1, 3], [0, 7]]]
    >>> shape_list(my_list_bis)
    NoConsistentError: La dimension 1 est non consistante.
    
    Notes
    -----
    La méthode vérifie la consistance de la liste. Si la liste n'est pas 
    consistante la fonction renvoit une erreur NoConsistentError.
    """
    if not isinstance(l, (list, tuple,)):
        return []
    if (isinstance(l[0], (list, tuple,))
            and len(set(map(lambda x: len(x), l))) > 1):
        raise NoConsistentError(
            "La dimension 0 est non consistante."
        )
    idx = 0
    return [len(l)] + _shape_list(l[0], idx)

def _shape_list(l: Union[List[Any], Tuple[Any]], idx: int):
    r"""
    Fonction privée. Permet de faire la récurcivité de shape_list tout en 
    indiquant la dimensions traité.
    
    Parametes :  Union[List[Any], Tuple[Any]]
        liste ou typle de dimension inconue
    idx : int
        indice de la dimensions pécédente
    """
    if not isinstance(l, (list, tuple,)):
        return []
    idx += 1
    if (isinstance(l[0], (list, tuple,))
            and len(set(map(lambda x: len(x), l))) > 1):
        raise NoConsistentError(
            f"La dimension {idx} est non consistante."
        )
    return [len(l)] + _shape_list(l[0], idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Modules callback.

Created on Sat Jan  9 10:42:31 2021

@author: Cyrile Delestre
"""

import os
import sys
import pickle as pkl
from abc import abstractmethod
from copy import deepcopy
from typing import List, Dict, Optional, Callable, Any, NewType
from dataclasses import dataclass, field

import numpy as np
from torch import Tensor, device as torch_device
from torch.cuda.amp import GradScaler
from tqdm import tqdm

from dstk.utils.meta_interface import SuperclassDocstring
from dstk.utils.errors import TrainError

# Permet d'éviter une initialisation circulaire via l'import de
# BaseEnvironnement
PyTorchEstimators = NewType('PyTorchEstimators', Any)

@dataclass
class FitState:
    r"""
    Classe contenant les informations sur l'apprentissage.

    Parameters
    ----------
    id_process : str
        Identifiant du processus.
    nb_epoch : int
        Nombre d'époques du processus d'apprentissage.
    epoch_max_steps : int
        Nombre d'itérations à l'intérieur d'une époque.
    eval_max_steps : int
        Nombre d'itérations à l'intérieur de la phase d'évaluation.
    loss_name : str
        Nom de la fonction coût.
    n_iter : int (=0)
        Numéro de l'itération en cours.
    n_epoch : int (=0)
        Numéro de l'époque en cours.
    """
    id_process: str
    nb_epoch: int
    epoch_max_steps: int
    eval_max_steps: int
    loss_name: str
    n_iter: int = 0
    n_epoch: int = 0

@dataclass
class FitControl:
    r"""
    Classe permettant de contôler le flux d'appretissage.

    Parameters
    ----------
    training_stop : bool (=False)
        Booléen indiquant si le processus d'apprentissage doit stopper ou non.
    diverge : bool (=False)
        Booléen indiquant si le processus d'apprentissage a divergé ou non.
    target_field: str (='target')
        Nom du champ de la target dans le dictionnaire d'échange.
    device : device (=device('cpu:0'))
        Device sur le quel le modèle exécute son entrainement.
    mixed_type_gpu : bool (=False)
        Indicateur pour que l'apprentissage utiliser les types mixé fp16 et 
        fp32 via l'API AMP (Automatic Mixed Precision).
    grad_scaler : GradScaler (=GradScaler(enabled=False))
    gradient_accumulation : Optional[int]
        Nombre d'itération pour l'accumulation du gradient. Si None ou < 1 
        mise du nombre d'itération à 1.
    max_gradient_norm : Optional[float]
        Valeur max du gradient pour son clipping.
    expected_key_to_device : List[str] (=list())
        Liste des variables à exclure vers la charge sur GPU.
    """
    training_stop: bool = False
    diverge: bool = False
    target_field: str = 'target'
    device: torch_device = torch_device('cpu:0')
    mixed_type_gpu: bool = False
    grad_scaler: GradScaler = GradScaler(enabled=False)
    gradient_accumulation: Optional[int] = None
    max_gradient_norm: Optional[float] = None
    expected_key_to_device: List[str] = field(default_factory=list)

    def __post_init__(self):
        if (
                self.gradient_accumulation is None
                or self.gradient_accumulation < 1
        ):
            self.gradient_accumulation = 1

class CallbackInterface:
    r"""
    Classe interface permettant de définir des modules utilisés durant 
    l'apprentissage de la méthode 
    :func:`~dstk.pytorch._base.BaseEnvironnement.fit` du module 
    :mod:`~dstk.pytorch._base.BaseEnvironnement`. Cette interface permet de 
    personnaliser facilement un callback.

    Notes
    -----
    Les méthodes qui sont appelées durant l'apprentissage :
        begin_train :
            au début de train : avant toute itération d'entrainement
        end_train :
            à la fin de train : après toute itération de récupération du 
            meilleur modèle
        begin_step :
            au début d'un batch d'apprentissage à chaque itération d'une 
            époque
        end_step :
            à la fin d'un batch d'apprentissage à chaque itération d'une 
            époque
        begin_epoch :
            au début d'une époque d'apprentissage
        end_epoch :
            à la fin d'une époque d'apprentissage
        begin_eval :
            au début d'une évaluation à la fin d'une époque
        end_eval :
            à la fin d'une évaluation à la fin d'une époque
        begin_eval_step :
            au début d'un batch d'évaluation à chaque itération d'une époque 
            d'évaluation
    """
    def copy(self):
        r"""
        Créer une copie de l'objet.
        """
        return deepcopy(self)

    @abstractmethod
    def begin_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: None,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée au début de train : avant toute itération 
        d'entrainement.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def end_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: None,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée à la fin de train : après toute itération de 
        récupération du meilleur modèle.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def begin_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée au début d'une époque d'apprentissage.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def end_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée à la fin d'une époque d'apprentissage.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def begin_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: None,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée au début d'une étape d'apprentissage.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def end_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée à la fin d'un étape d'apprentissage.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def begin_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: None,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée au début d'une évaluation à la fin d'une époque.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée à la fin d'un évaluation à la fin d'une époque.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def begin_eval_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée au début d'une étape de validation.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        """
        pass

    @abstractmethod
    def error_interrupt(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        exception: Exception,
        **kwargs
    ) -> None:
        r"""
        Méthode appelée si une erreur a été rencontré.

        Parameters
        ----------
        model : PyTorchEstimators
            Référence au modèle sur lequel l'apprentissage se déroule.
        data : Iterable
            Data sous forme de liste ou d'un DataLoader.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        res_loss : float
            Valeur de la fonction coût.
        exeption: Exception
            Type de l'exeption renvoyé.
        """
        pass

class CallbackHandler(CallbackInterface):
    r"""
    Classe handler permettant de gérer simplement une liste de callbacks.
    """
    def __init__(self, callbacks: Optional[List[CallbackInterface]]):
        type_cb = list(
            map(lambda x: isinstance(x, CallbackInterface), callbacks)
        )
        if not all(type_cb):
            list_name = [
                cb.__class__.__name__ for ii, cb in enumerate(callbacks)
                if not type_cb[ii]
            ]
            raise AttributeError(
                f"Les callbacks {', '.join(list_name)} ne sont pas de type "
                "CallbackInterface."
            )
        order = [
            0 if cc.__class__.__name__=='ProgressBarCallback' else ii
            for ii, cc in zip(range(1, len(callbacks)+1), callbacks)
        ]
        callbacks = list(
            map(
                lambda x: x[1],
                sorted(zip(order, callbacks), key=lambda x: x[0])
            )
        )
        self.callbacks = callbacks

    def copy(self):
        self.callbacks = [cb.copy() for cb in self.callbacks]

    def begin_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call(
            step='begin_train',
            model=model,
            data=data,
            state=state,
            control=control,
            res_loss=res_loss,
            **kwargs
        )

    def end_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call('end_train', model, data, state, control, res_loss, **kwargs)

    def begin_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call(
            step='begin_epoch',
            model=model,
            data=data,
            state=state,
            control=control,
            res_loss=res_loss,
            **kwargs
        )

    def end_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call('end_epoch', model, data, state, control, res_loss, **kwargs)

    def begin_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call(
            step='begin_step',
            model=model,
            data=data,
            state=state,
            control=control,
            res_loss=res_loss, 
            **kwargs
        )

    def end_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call('end_step', model, data, state, control, res_loss, **kwargs)

    def begin_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call(
            step='begin_eval',
            model=model,
            data=data,
            state=state,
            control=control,
            res_loss=res_loss,
            **kwargs
        )

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call('end_eval', model, data, state, control, res_loss, **kwargs)

    def begin_eval_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.call(
            step='begin_eval_step',
            model=model,
            data=data,
            state=state,
            control=control,
            res_loss=res_loss,
            **kwargs
        )

    def error_interrupt(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        exception: Exception,
        **kwargs
    ) -> None:
        self.call(
            step='error_interrupt',
            model=model,
            data=data,
            state=state,
            control=control, 
            res_loss=res_loss,
            exception=exception,
            **kwargs
        )

    def call(
        self,
        step: str,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        for cb in self.callbacks:
            getattr(cb, step)(model, data, state, control, res_loss, **kwargs);

@dataclass
class PrintCallback(CallbackInterface):
    r"""
    Callback simple permettant d'afficher la performance du modèle en cours 
    d'apprentissage après l'évaliation.
    """
    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        print(
            f"\nid : {state.id_process}\n"
            f"Epoch : {state.n_epoch+1}/{state.nb_epoch}, "
            f"{state.loss_name} : {res_loss:.4f}\n",
            flush=True
        )

    def error_interrupt(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        exception: Exception,
        **kwargs
    ) -> None:
        TrainError(
            f"Apprentissage arrêté du à l'error: {exception}"
        )

@dataclass
class Backup(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Callback de backup de l'apprentissage régulier à chaque intervale. Cette 
    sauvegarde préserve l'état des poids ainsi que l'état de l'optimizer. Il 
    est alors possible de relancer l'apprentissage dans l'état du système 
    d'apprentissage au moment de l'arrêt imprévue. La sauvegarde ca fait 
    au format Pickle avec le schéma suivant :
        parameters.stat_dict, optimizer.stat_dict, state, control

    Parameters
    ----------
    path_backup : Optional[str]
        Chemin de backup. Si None utilise le répertoire dans le quel est 
        lancé l'instance Python. Le nom du backup est l'id_process.
    n_iter : int (=1_000)
        Nombre d'itération entre deux backup.
    """
    path_backup: Optional[str] = None
    n_iter: int = 1_000

    def __post_init__(self):
        self.path_backup = (
            self.path_backup if self.path_backup else os.getcwd()
        )

    def end_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if state.n_iter % self.n_iter == 0:
            pkl.dump(
                [
                    model.state_dict(),
                    model.optimizer.state_dict(),
                    state,
                    control
                ],
                open(f"{self.path_backup}/{state.id_process}.bck", "wb")
            )

    def end_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        os.remove(f"{self.path_backup}/{state.id_process}.bck")

    def error_interrupt(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        exception: Exception,
        **kwargs
    ) -> None:
        if isinstance(exception, KeyboardInterrupt):
            os.remove(f"{self.path_backup}/{state.id_process}.bck")

@dataclass
class EarlyStoppingCallback(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Callback de gestion du early-stopping. Permet de sauvegarder la meilleure 
    itération durant l'apprentissage. Si un early_stopping_rounds est donné 
    alors l'apprentissage se stoppe si la performance n'a pas été meilleure au 
    bout de ces early_stopping_rounds itérations.

    Parameters
    ----------
    path_early_stopping : Optional[str] (=None)
        Chemin (absolu ou relatif) où le modèle par early stopping sera 
        préservé. Si None, les paramètres seront sauvegardés en RAM. 
        Cependant il est déconseillé de le faire car les réseaux profonds 
        peuvent être très gourmands en espace RAM. De même, en cas de 
        plantage, un état du réseau pouvant être satisfaisant est préservé 
        sur le disque dur.
    early_stopping_rounds : Optional[int] (=None)
        Nombre d'itérations (epoch) sans amélioration. Si None pas de 
        early stopping, le modèle sauvegardé sera donc le meilleur modèle 
        durant l'ensemble des époques.
    minimize : bool (=True)
        Indique si l'objectif est de minimiser la fonciton coût (True) ou 
        de la maximiser (False).
    verbose : bool (=False)
        Indique les informations de gain de performance.
    """
    path_early_stopping: Optional[str] = None
    early_stopping_rounds: Optional[int] = None
    minimize: bool = True
    verbose: bool = False

    def __post_init__(self):
        self.old_res_loss = (np.inf if self.minimize else -np.inf)
        self.iter_stop: int = 0
        self.best_epoch: int = 0
        self.best_iter: int = 0

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if (
                res_loss < self.old_res_loss 
                if self.minimize else 
                res_loss > self.old_res_loss
        ):
            if self.verbose:
                print(
                    "\nGain de performance époque/itération : "
                    f"{state.n_epoch+1}/{state.n_iter+1} ;\n"
                    f"Gain de performance {state.loss_name} :\n"
                    f"{self.old_res_loss} -> {res_loss}.\n",
                    flush=True
                )
            self.iter_stop = 0
            self.old_res_loss = res_loss
            self.best_epoch = state.n_epoch
            self.best_iter = state.n_iter
            if self.path_early_stopping is None:
                self.state_dict = model._extract_state(model)
            else:
                model.save_model(
                    f"{self.path_early_stopping}/{state.id_process}.model"
                )
        elif (
                self.early_stopping_rounds is not None and 
                (
                    res_loss >= self.old_res_loss 
                    if self.minimize else
                    res_loss <= self.old_res_loss
                )
        ):
            self.iter_stop += 1
            if self.iter_stop >= self.early_stopping_rounds:
                control.training_stop = True

    def end_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.verbose:
            print(
                f"\nMeilleure époque/iteration : {self.best_epoch+1}/"
                f"{self.best_iter+1}\n"
                f"Meilleure performance {state.loss_name} : "
                f"{self.old_res_loss}.\n",
                flush=True
            )
        if self.path_early_stopping is None:
            model.load_state_dict(self.state_dict)
        else:
            model.load_weights(
                f"{self.path_early_stopping}/{state.id_process}.model"
            )
            os.remove(f"{self.path_early_stopping}/{state.id_process}.model")

    def error_interrupt(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        exception: Exception,
        **kwargs
    ) -> None:
        if isinstance(exception, KeyboardInterrupt):
            print(
                "\nInterruption par la commande KeyboardInterrupt.",
                flush=True
            )
            self.end_train(
                model=model,
                data=data,
                state=state,
                control=control,
                res_loss=res_loss,
                **kwargs
            )

@dataclass
class LRSchedulerCallback(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Classe de gestion du learning rate. Ce callback ne met à jour le 
    learning rate que durant les périodes d'évaluation.

    Parameters
    ----------
    lr_scheduler : Callable
        Classe non instanciée d'un planificateur de learning rate. Les 
        planificateurs de learning rate se situent dans 
        torch.optim.lr_scheduler et permettent de contrôler le lerning rate au 
        travers des époques.
    scheduler_kwargs : Dict[str, Any] (=dict())
        Argument d'entrée de scheduler (en dehors de l'optimizer qui doit 
        être inclu à l'intérieur de la classe du réseau). Explicitement 
        l'instanciation sera faite dans la méthode fit sera du type :
            lr_scheduler(self.optimizer, **scheduler_kwargs)
    by_step : bool (=False)
        Détermine si le LRScheduler est appelé à la fin de la phase 
        d'évaluation (=False) ou à la fin de chaque step de train (=True).
    verbose : bool (=True)
        Permet d'affichier le lerning rate initial et les évolutions du 
        learning rate quand ils ont lieu. Par défaut cette argument est 
        à True car la sortie n'est pas très verbeuse.

    Notes
    -----
    Si l'entrainement ne possède pas de dataset d'évaluation aucune mise à 
    jour de learning rate ne sera fait durant l'entrainement par 
    l'intermédiaire de ce callback.

    Si le scheduler est éxécuté à chaque étape de l'entrainement 
    (by_step=True) le lr_scheduler ne peut pas être ReduceLROnPlateau.
    """
    lr_scheduler: Callable
    scheduler_kwargs: Dict[str, Any] = field(default_factory=dict)
    by_step: bool = False
    verbose: bool = True

    def __post_init__(self):
        if self.lr_scheduler.__name__ == "ReduceLROnPlateau" and self.by_step:
            raise AttributeError(
                "Le lr_scheduler est le ReduceLROnPlateau et ne peut pas "
                "être appliqué durant la phase d'entrainement, mais seulement "
                "durant la phase d'évaluation."
            )

    def begin_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.scheduler = self.lr_scheduler(
            model.optimizer,
            **self.scheduler_kwargs
        )
        if self.verbose:
            self.old_lr = model.optimizer.state_dict()['param_groups'][0]['lr']
            print(
                f"A l'initialisation le lr = {self.old_lr}\n",
                flush=True
            )

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if not self.by_step:
            if self.lr_scheduler.__name__ == "ReduceLROnPlateau":
                self.scheduler.step(res_loss)
            else:
                self.scheduler.step()
            if self.verbose:
                lr = model.optimizer.state_dict()['param_groups'][0]['lr']
                if self.old_lr != lr:
                    print(
                        f"\nA l'époque {state.n_epoch+1} itération "
                        f"{state.n_iter+1} le learning rate a évolué :\n"
                        f"{self.old_lr} -> {lr}.\n",
                        flush=True
                    )
                    self.old_lr = lr

    def end_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.by_step:
            if self.lr_scheduler.__name__ == "ReduceLROnPlateau":
                self.scheduler.step(res_loss)
            else:
                self.scheduler.step()
            if self.verbose:
                lr = model.optimizer.state_dict()['param_groups'][0]['lr']
                if self.old_lr != lr:
                    print(
                        f"\nA l'époque {state.n_epoch+1} itération "
                        f"{state.n_iter+1} le learning rate a évolué :\n"
                        f"{self.old_lr} -> {lr}.\n",
                        flush=True
                    )
                    self.old_lr = lr

@dataclass
class ProgressBarCallback(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Barre de progression TQDM sur les époques et les évaluations.

    mininterval : float
        Temps en seconde entre les intervalles de raffraichissement de la 
        barre de progression. Par défaut 0.5.
    bar_epoch : bool
        Indication si la barre de progression sera exécutée sur une époque 
        d'apprentissage durant l'entrainement. Par défaut True.
    bar_eval : bool
        Indication si la barre de progression sera exécutée sur une époque 
        d'évaluation durant l'entrainement. Par défaut False.
    """
    mininterval: float = 0.5
    bar_epoch: bool = True
    bar_eval: bool = False
    _swa_process: bool = False

    def begin_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.bar_epoch:
            self.epoch_bar = tqdm(
                total=state.epoch_max_steps,
                desc=(
                    f"Epoque {state.n_epoch_swa+1}/{state.nb_epoch_swa}"
                    if self._swa_process else
                    f"Epoque {state.n_epoch+1}/{state.nb_epoch}"
                ),
                mininterval=self.mininterval,
                leave=True,
                file=sys.stdout
            )

    def end_epoch(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.bar_epoch:
            self.epoch_bar.refresh()
            self.epoch_bar.close()

    def begin_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.bar_epoch:
            self.epoch_bar.update(1)

    def begin_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: None,
        **kwargs
    ) -> None:
        if self.bar_eval:
            self.eval_bar = tqdm(
                total=state.eval_max_steps,
                desc="Evaluation",
                mininterval=self.mininterval
            )

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.bar_eval:
            self.epoch_bar.refresh()
            self.eval_bar.close()

    def begin_eval_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        if self.bar_eval:
            self.eval_bar.update(1)

class Supervision(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Callback permettant de faire le suivit de l'apprentissage via Tensorboard.
    Il indique l'évolution de la fonciton coût sur la base d'évaluation et
    d'entrainement.

    Parameters
    ----------
    log_dir : Optional[str]
        Chemin de sauvegarde des données Tensorboard, par défaut 
        runs/CURRENT_DATETIME_HOSTNAME.
    comment : str (='')
        Concatène un commentaire au log_dir. Si log_dir est renseigné cette 
        argument n'est pas pris en compte.
    purge_step: Optional[int]
        When logging crashes at step T+X and restarts at step T, any 
        events whose global_step larger or equal to T will be purged and 
        hidden from TensorBoard. Note that crashed and resumed experiments 
        should have the same log_dir.
    max_queue : int (=10)
        Taille de la queue en attente avec de mettre sur le disque.
    flush_secs : int (=120)
        Temps entre deux mise sur le disque.
    filename_suffix : str (='')
        Ajout d'un suffix au log_dir.
    """
    def __init__(
        self,
        log_dir: Optional[str]=None,
        comment: str='',
        purge_step: Optional[int]=None,
        max_queue: int=10,
        flush_secs: int=120,
        filename_suffix: str=''
    ):
        try:
            global SummaryWriter
            from torch.utils.tensorboard import SummaryWriter
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                    "Le module TensorBoard n'est pas installé : "
                    "l'instruction  torch.utils.tensorboard renvoit une "
                    "erreur ModuleNotFoundError. Visitez la page du projet "
                    "PyTorch pour savoir comment installer le package "
                    "TorchVision qui comprend TensorBoard à destination de "
                    "PyTorch. Ou installez directement TensorBoard via pip "
                    "ou conda."
                )
        self.writer = SummaryWriter(
            log_dir=log_dir,
            comment=comment,
            purge_step=purge_step,
            max_queue=max_queue,
            flush_secs=flush_secs,
            filename_suffix=filename_suffix
        )

    def end_step(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.writer.add_scalar(
            'Loss/train',
            res_loss,
            state.n_iter
        )
        self.writer.add_scalar(
            'LR/lr',
            model.optimizer.param_groups[0]['lr'],
            state.n_iter
        )

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        **kwargs
    ) -> None:
        self.writer.add_scalar('Loss/eval', res_loss, state.n_iter)

DEFAULT_CALLBACKS = [
    ProgressBarCallback(),
    PrintCallback(),
    EarlyStoppingCallback(verbose=True),
    Backup()
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe RocAnalyser permettant de calculer, mettre en forme simplement des 
courbes ROC.

Created on Tue Nov 24 12:50:56 2020

@author: Cyrile Delestre
"""

from dataclasses import dataclass, field
from typing import List, Optional
from itertools import product, chain
from collections import OrderedDict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, auc, roc_auc_score
from joblib import Parallel, delayed
from tqdm import tqdm

from .statistique import cdf_normal, cdf_rayleigh, kolmogorov_smirnov_test
from dstk.utils import weighted_avg_and_std, chunker

@dataclass
class ROCAnalyzer:
    r"""
    Classe d'analyse d'un modèle de classification multi-classe utilisant les 
    propriétés de la courbe ROC (Receiver Operating Characteristic). La 
    classe comporte également un gestionnaire des plots pour faciliter 
    l'affichage des figures.
    
    Parameters
    ----------
    y_true : np.ndarray
        ndarray de dimension (n_samples,) où chaque classe est représentée par 
        un int. L'entier doit correspondre à l'argument de la classe (0 pour 
        la première classe, 1 pour la seconde, etc.). Si ce n'est pas le cas 
        faire :
            np.argmax(y_true)
    y_score : np.ndarray
        ndarray de dimension (n_samples, n_classes) contenant les probabilités 
        d'appartenance de chaque classe. C'est typiquement la sortie de type 
        Scikit-Learn model.predict_propba(X_eval). Attention le cas bi-classe 
        y_score doit être de dimension (n_samples, 2).
    labels : Optional[List[str]]
        nom des labels du modèle. Doit suivre l'ordre de l'indexation de 
        y_true. Si non-renseigné (None) utilise l'indexation de y_true pour 
        identifier les classes.
    sample_weight : Optional[np.ndarray]
        poids sur les samples.
    drop_intermediate : bool
        (paramètre de sklearn.metrics.roc_curve), drop un threshold 
        sous-optimal (qui n'apparait pas sur la courbe car trop rapproché 
        d'un autre). Permet d'alléger la représentation.
    densities : List[str]
        liste des densités théoriques à tester pour l'estimation de courbe 
        ROC théorique. Attention certaines peuvent demander le package SciPy. 
        Le choix pour déterminer la meilleure loi adaptée pour la 
        représentation de la densité du score sera la loi ayant l'écart 
        minimum de la différence de Kolmogorov-Smirnov. Par défaut seule la 
        loi Normale est utilisée, on appelle cette estimation de la courbe ROC 
        la courbe ROC Binormale. Liste possible implanté pour le moment :
            'normal', 'rayleigh'
    roc_multi_class : str
        type de courbe ROC dans le cadre d'une modélisation multi-classe. 
        Soit 'one-vs-rest' (par défaut) et chaques classes seront 
        repsésentées. Dans ce cas l'AUC sera la probabilité que le score de 
        la classe i soit supérieur au score de la classe non i. Sinon 
        'one-vs-one' où on test la classe i sachant qu'on est dans la classe 
        j. Dans ce cas  l'AUC sera la probabilité que le score de la classe i 
        soit supérieur à classe réelle j.

    Notes
    -----
    La partie estimation de la courbe ROC théorique est encore en cours de 
    réflexion...

    Les attributs liés à l'analyse ne sont pas générés tant que l'analyse 
    en question n'a pas tourné. Par exemple on a pas les moyennes des classes 
    tant que 
    :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.process_roc_theo` 
    ne s'est pas exécuté. Voir dans la doc de chacunes des méthodes 
    (généralement commençant par ``process_``) les attributs d'analyse qui y 
    sont générés.
    
    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.datasets import make_moons
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.model_selection import train_test_split
    >>> from dstk.visualization import ROCAnalyser
    >>> 
    >>> # Exemple de données d'une classification binaire généré
    >>> # procéduralement
    >>> data_moons, target_moons = make_moons(
    >>>     n_samples = 1000,
    >>>     shuffle = True,
    >>>     noise = 0.5,
    >>>     random_state = 42
    >>> )
    >>> 
    >>> (X_train_moons, X_eval_moons, y_train_moons,
    >>> y_eval_moons) = train_test_split(
    >>>     data_moons,
    >>>     target_moons,
    >>>     train_size = 800,
    >>>     random_state = 42
    >>> )
    >>> 
    >>> rf = RandomForestClassifier(random_state = 42)
    >>> rf.fit(X_train_moons, y_train_moons)
    >>> 
    >>> y_score_moons = rf.predict_proba(X_eval_moons)
    >>> 
    >>> ana_moons = ROCAnalyser(
    >>>     y_eval_moons,
    >>>     y_score_moons,
    >>>     densities = ['normal', 'rayleigh']
    >>> )
    >>> ana_moons.plot_roc_bootstrap().plot_roc_theo().plot_roc()
    
    Affiche la courbe ROC de ce cas bi-classe et affiche également 
    l'estimation de la courbe théorique de ce modèle ainsi la précision de 
    la courbe ROC via une approche de bootstraping.
    
    >>> analyse = ana_moons.get_annalyse()
    >>> pd.DataFrame(analyse)
           best_tresh  roc_auc  best_tresh_theo ... 1_ks_normal  1_ks_rayleigh
        0         0.4  0.88055           0.4987 ...    0.148793       0.909485
    
    Voici un autre exemple d'un modèle 3 classes sur un dataset généré.
    
    >>> from sklearn.datasets import make_multilabel_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from dstk.visualization import ROCAnalyser
    >>> 
    >>> data_wine, target_wine = make_multilabel_classification(
    >>>     n_samples = 1000,
    >>>     n_features = 5,
    >>>     n_classes = 3,
    >>>     random_state = 42
    >>> )
    >>> 
    >>> (X_train_wine, X_eval_wine, y_train_wine,
    >>> y_eval_wine) = train_test_split(
    >>>     data_wine,
    >>>     np.argmax(target_wine, axis = 1),
    >>>     train_size = 25,
    >>>     random_state = 42
    >>> )
    >>> 
    >>> lr = LogisticRegression(max_iter = 1e3, random_state = 42)
    >>> lr.fit(X_train_wine, y_train_wine)
    >>> 
    >>> y_score_wine = lr.predict_proba(X_eval_wine)
    >>> ana_wine = ROCAnalyser(
    >>>     y_eval_wine,
    >>>     y_score_wine,
    >>>     labels = ['Champagne', 'Bordeau', 'Beaujolais'],
    >>>     densities = ['normal', 'rayleigh']
    >>> )
    >>> ana_wine.plot_roc_manager(3).plot_roc().plot_roc_theo()
    
    Affiche les courbes ROC associées à chaques classes ainsi que les
    estimations des courbes théoriques de ce modèle.
    
    >>> ana_wine.process_performance_matrix()
                    Champagne   Bordeau  Beaujolais
        Champagne    0.954217  0.207498    0.233000
        Bordeau      0.190446  0.916519    0.577172
        Beaujolais   0.052897  0.334532    0.981586
    """
    y_true: np.ndarray
    y_score: np.ndarray
    labels: Optional[List[str]]=None
    sample_weight: Optional[np.ndarray]=None
    drop_intermediate: bool=True
    densities: List[str]=field(default_factory = lambda: ['normal'])
    roc_multi_class: str='one-vs-one'
    
    def __post_init__(self):
        self.nb_class = self.y_score.shape[1]
        if self.nb_class == 1:
            raise AttributeError(
                "Il n'y a qu'une seule classe, y_score doit être de "
                "dimension : (n_samples, n_classes)."
            )
        if not self.labels is None and len(self.labels) != self.nb_class:
            raise AttributeError(
                "Il n'y a pas le bon nombre d'éléments dans labels. Le "
                f"nombre d'éléments attendu est {self.nb_class} alors que "
                f"labels a une taille de {len(self.labels)}. Pour rappel "
                "labels peut être None."
            )
        if not self.roc_multi_class in ['one-vs-one', 'one-vs-rest']:
            raise AttributeError(
                "L'argument roc_multi_class doit être soit : 'one-vs-one' "
                f"ou 'one-vs-rest' et non '{self.roc_multi_class}'."
            )
        self.analyse = OrderedDict()
        fn_density = dict(
            normal = cdf_normal,
            rayleigh = cdf_rayleigh
        )
        self.list_density = [
            (name, fn_density[name]) for name in self.densities
        ]

    def _roc(self, ratio):
        r"""
        Méthode privée calculant la courbe ROC, l'AUC et d'autres attributs.
        """
        (fpr, tpr, thresh, best_thresh, best_fpr, best_tpr, 
         roc_auc, nb_obs, idx_sel) = [], [], [], [], [], [], [], [], []
        size  = len(self.y_true)
        idx = np.sort(
            np.random.choice(
                range(size),
                size = int(ratio*size),
                replace = False
            )
        )
        if self.nb_class == 2:
            fpr_, tpr_, thresh_ = roc_curve(
                y_true = self.y_true[idx],
                y_score = self.y_score[idx, 1],
                sample_weight = self.sample_weight,
                drop_intermediate = self.drop_intermediate
            )
            idx_best = np.argmax(tpr_ - fpr_)
            fpr.append(fpr_)
            tpr.append(tpr_)
            thresh.append(thresh_)
            best_thresh.append(thresh_[idx_best])
            best_fpr.append(fpr_[idx_best])
            best_tpr.append(tpr_[idx_best])
            roc_auc.append(auc(fpr_, tpr_))
            nb_obs.append(np.sum(self.y_true[idx] == 2))
            idx_sel.append(idx)
        else:
            for ii in range(self.nb_class):
                if self.roc_multi_class == 'one-vs-rest':
                    y_score = self.y_score[idx, ii]
                elif self.roc_multi_class == 'one-vs-one':
                    filt_not_class = ~(self.y_true[idx]==ii).ravel()
                    y_score = self.y_score[idx, ii]
                    y_score[filt_not_class] = (
                        (1-self.y_score[idx, self.y_true[idx].ravel()])
                        [filt_not_class]
                    )
                else:
                    raise AttributeError(
                        "L'argument roc_multi_class doit être soit : "
                        "'one-vs-one' ou 'one-vs-rest' et non "
                        f"'{self.roc_multi_class}'."
                    )
                fpr_, tpr_, thresh_ = roc_curve(
                    y_true = (self.y_true[idx] == ii).astype(int),
                    y_score = y_score,
                    sample_weight = self.sample_weight,
                    drop_intermediate = self.drop_intermediate
                )
                idx_best = np.argmax(tpr_ - fpr_)
                fpr.append(fpr_)
                tpr.append(tpr_)
                thresh.append(thresh_)
                best_thresh.append(thresh_[idx_best])
                best_fpr.append(fpr_[idx_best])
                best_tpr.append(tpr_[idx_best])
                roc_auc.append(auc(fpr_, tpr_))
                nb_obs.append(np.sum(self.y_true[idx] == ii))
                idx_sel.append(idx)
        return list(
            zip(fpr, tpr, thresh, best_thresh, best_fpr,
                best_tpr, roc_auc, nb_obs, idx_sel)
        )

    def process_roc(self):
        r"""
        Méthode de calcul de la courbe ROC pour chacune des classes. Cette 
        méthode ne peut être appliquée qu'une seule fois car son résultat est 
        stocké dans l'attribut 'data_roc'.
        
        Notes
        -----
        Dans le cadre de l'attribut 'analyse' certains de ses éléments sont 
        calculés ici :
        
        'best_tresh' :
            correspondant au seuil optimal de modèle au sens de la courbe ROC.
        'roc_auc' :
            AUC associé à la classe.
        'nb_obs' :
            Nombre d'observation par classe.
        
        See also
        --------
        process_roc_theo, process_roc_bootstrap, process_performance_matrix
        """
        self.data_roc = self._roc(ratio=1)
        data_roc = list(zip(*self.data_roc))
        self.analyse['best_thresh'] = data_roc[3]
        self.analyse['roc_auc'] = data_roc[6]
        self.analyse['nb_obs'] = data_roc[7]
        return self

    def process_roc_bootstrap(self,
                              monte_carlo: int=100,
                              ratio: float=0.8,
                              n_jobs: int=1,
                              verbose: bool=False):
        r"""
        Méthode d'estimation de la courbe ROC et de ses statistiques par 
        bootstrap. Permet d'apprécier la volatilité de l'estimateur et de 
        donner un AUC ou un seuillage optimal plus robuste.
        
        Parameters
        ----------
        :monte_carlo: nombre de courbe ROC estimées (par défaut 100).
        :ratio: ratio d'observation conservé pour une estimation de la courbe 
            ROC (par défaut 0.8).
        :n_jobs: permet de paralléliser les calculs des courbes ROC et des 
            AUC.
        :verbose: permet d'afficher une barre d'avancement.
        
        Notes
        -----
        Dans le cadre de l'attribut 'analyse' certains de ses éléments sont 
        calculés ici :
        
        'best_thresh_bootstrap_mean' :
            correspondant au seuil moyen optimal du modèle au sens de la 
            courbe ROC estimée.
        'best_thresh_bootstrap_med' :
            correspondant au seuil median optimal du modèle au sens de la 
            courbe ROC estimée.
        'best_thresh_bootstrap_std' :
            correspondant à l'écart-type du seuil optimal du modèle au sens 
            de la courbe ROC estimée.
        'roc_auc_booststrap_mean' :
            AUC moyen associé à la classe déterminée sur l'estimation de la 
            courbe ROC.
        'roc_auc_booststrap_med' :
            AUC meidan associé à la classe déterminée sur l'estimation de la 
            courbe ROC.
        'roc_auc_booststrap_std' :
            écart-type de l'AUC associé à la classe déterminée sur 
            l'estimation de la courbe ROC.
        'nb_obs_booststrap' :
            nombre d'observation pour l'estimation de la courbe ROC.
        
        See also
        --------
        process_roc, process_roc_theo, process_performance_matrix
        """
        iterator = range(monte_carlo)
        if verbose:
            iterator = tqdm(
                iterator,
                desc = f'{self.process_roc_bootstrap.__name__}'
            )
        with Parallel(n_jobs=n_jobs) as par:
            self.data_roc_bootstrap = par(
                delayed(self._roc)(ratio) for ii in iterator
            )
        data_roc_bootstrap = list(zip(*self.data_roc_bootstrap))
        best_tresh = list(
            map(lambda x: list(map(lambda y: y[3], x)), data_roc_bootstrap)
        )
        roc_auc = list(
            map(lambda x: list(map(lambda y: y[6], x)), data_roc_bootstrap)
        )
        self.analyse['best_thresh_bootstrap_mean'] = (
            np.mean(best_tresh, axis=1).tolist()
        )
        self.analyse['best_threch_bootstrap_med'] = (
            np.median(best_tresh, axis=1).tolist()
        )
        self.analyse['best_thresh_bootstrap_std'] = (
            np.std(best_tresh, axis=1).tolist()
        )
        self.analyse['roc_auc_bootstrap_mean'] = (
            np.mean(roc_auc, axis=1).tolist()
        )
        self.analyse['roc_auc_bootstrap_med'] = (
            np.median(roc_auc, axis=1).tolist()
        )
        self.analyse['roc_auc_bootstrap_std'] = (
            np.std(roc_auc, axis=1).tolist()
        )
        self.analyse['nb_obs_bootstrap'] = list(
            map(lambda x: list(map(lambda y: y[7], x))[0], data_roc_bootstrap)
        )
        return self

    def process_roc_theo(self):
        r"""
        Méthode déterminant la densité théorique la plus proche de la 
        densité du score et calculant la courbe ROC théorique associée aux 
        densités théoriques. La liste de choix des lois théoriques est 
        initiée à l'instanciation de la classe ROCAnalyser dans l'attribut 
        'densities'.
        
        Notes
        -----
        La courbe ROC étant par définition dans
        Scikite-Learn par :
            fpr, tpr, thresh = ROC(y_true, y_socre)
        
        Si on détermine la loi de répartition des faux positifs et des vrais 
        positifs alors :
            fpr, tpr, thresh = CDF(law_fp, thresh), CDF(law_tp, thresh), thresh
        
        D'autres approches sont possibles mais nécessitent obligatoirement 
        SciPy (ce qu'on ne souhaite pas) car il faut pouvoir calculer CDP^-1.
        
        Seules les lois Normale, Rayleigh ne nécessitent pas le package SciPy.
        
        Dans le cadre de l'attribut 'analyse' certains de ses éléments sont 
        calculés ici :
        
        'best_thresh_theo' :
            correspondant au seuil optimal de modèle au sens de la courbe ROC 
            théorique.
        'roc_auc_theo' :
            AUC associé à la classe déterminée sur la courbe ROC théorique.
        mean_{classe_i} :
            moyennes des classes (vrai positive et fausses positives) pour 
            chaque hypothèse d'appartenance.
        std_{classe_i} :
            écart-types des classes (vrai positive et fausses positives) pour 
            chaque hypothèse d'appartenance.
        {classe_i}_ks_{densité_j} :
            écart de Kolmogorov-Smirnov de la classe i pour la densité j pour 
            chaque hypothèse.
        
        See also
        --------
        process_roc, process_roc_bootstrap, process_performance_matrix
        """
        thresh_theo_ = np.linspace(-6, 7, 5000)
        bins = [-6, -4, -1]+np.arange(0,1.1,0.1).tolist()+[1, 2, 5, 7]
        (fpr_theo, tpr_theo, thresh_theo, best_thresh_theo, best_fpr_theo, 
         best_tpr_theo, roc_auc_theo) = [], [], [], [], [], [], []
        mean_over_classes, std_over_classes, ks_over_classes = [], [], []
        for ii in range(self.nb_class):
            if self.nb_class == 2:
                ii = 1
            y_ = [
                self.y_score[self.y_true == cc, ii]
                for cc in range(self.nb_class)
            ]
            mean, std = list(
                zip(
                    *[weighted_avg_and_std(
                        x = yy,
                        sample_weight = self.sample_weight,
                        axis = 0
                    ) for yy in y_]
                )
            )
            mean, std = np.array(mean).ravel(), np.array(std).ravel()
            dens_select = []
            ks_ov = []
            for qq in range(self.nb_class):
                ks = []
                for name, cdf in self.list_density:
                    ks_ = kolmogorov_smirnov_test(
                        x = y_[qq],
                        cdf = cdf,
                        mean = mean[qq],
                        std = std[qq],
                        bins = bins
                    )
                    ks.append((cdf, ks_))
                ks_sort = sorted(ks, key=lambda x: x[1], reverse=False)
                dens_select.append(ks_sort[0][0])
                ks_ov.append([kk for _, kk  in ks])
            tpr_ = 1-dens_select[ii](
                thresh_theo_,
                mean = mean[ii],
                std = std[ii]
            )
            dens_select_other = dens_select
            dens_select_other.pop(ii)
            mean_other = np.delete(mean, ii)
            std_other = np.delete(std, ii)
            fpr_ = np.zeros_like(thresh_theo_)
            for qq, cdf in enumerate(dens_select_other):
                fpr_ += 1-cdf(
                    thresh_theo_,
                    mean = mean_other[qq],
                    std = std_other[qq]
                )
            fpr_ /= len(dens_select_other)
            mean_over_classes.append(mean)
            std_over_classes.append(std)
            ks_over_classes.append(ks_ov)
            idx = np.argmax(tpr_ - fpr_)
            fpr_theo.append(fpr_)
            tpr_theo.append(tpr_)
            thresh_theo.append(thresh_theo_)
            best_thresh_theo.append(thresh_theo_[idx])
            best_fpr_theo.append(fpr_[idx])
            best_tpr_theo.append(tpr_[idx])
            roc_auc_theo.append(auc(fpr_, tpr_))
            if self.nb_class == 2:
                break
        self.data_roc_theo = list(
            zip(
                fpr_theo,
                tpr_theo,
                thresh_theo,
                best_thresh_theo,
                best_fpr_theo,
                best_tpr_theo,
                roc_auc_theo
            )
        )
        self.analyse['best_thresh_theo'] = best_thresh_theo
        self.analyse['roc_auc_theo'] = roc_auc_theo
        mean_unzip = list(zip(*mean_over_classes))
        sdt_unzip = list(zip(*std_over_classes))
        ks_unzip_ = list(zip(*ks_over_classes))
        for ii in range(self.nb_class):
            self.analyse[
                f'mean_{ii if self.labels is None else self.labels[ii]}'
            ] = mean_unzip[ii]
            self.analyse[
                f'std_{ii if self.labels is None else self.labels[ii]}'
            ] = sdt_unzip[ii]
            ks_unzip = list(zip(*ks_unzip_[ii]))
            for qq, nn in enumerate(self.list_density):
                mm, _ = nn
                self.analyse[
                    f'{ii if self.labels is None else self.labels[ii]}_ks_{mm}'
                ] = ks_unzip[qq]
        return self

    def _auc_par(self, lot):
        r"""
        Méthode interne permettant de calculer en parallèle des AUC.
        """
        auc_res = []
        for class_a, class_b in lot:
            auc_res.append(
                roc_auc_score(
                    (self.y_true == class_a).astype(int),
                    self.y_score[:, class_b],
                    sample_weight = self.sample_weight
                )
            )
        return auc_res

    def process_performance_matrix(self,
                                   size: int=10,
                                   n_jobs: int=1,
                                   verbose: bool=False):
        r"""
        Méthode de calcul de la matrice de performance au sens du score 
        d'un modèle de classification.
        
        Parameters
        ----------
        size : int
            taille des tronçons où les combinaisons des classes seront mises. 
            Surtout utile si n_jobs > 1.
        n_jobs : int
            permet de paralléliser le calcul des AUC sur toutes les 
            combinaisons des classes (nombre de calcul d'AUC == nb_class**2).
            Pas très utile si le nombre de classe est faible.
        verbose : bool
            permet d'afficher une barre d'avancement des calculs des AUC au 
            travers des différents tronçons.
        
        Returns
        -------
        DataFrame contenant la matrice de performance où les noms des colonnes 
        et indices font référence aux noms des classes. Le résultat est 
        stocké dans l'attribut 'performance_matrix' (pas besoin de calculer 
        tout le temps cette matrice).
        
        Notes
        -----
        Si on note S(X_i) un score de la classe i et H_i l'hypothèse qu'on 
        est dans la classe i, alors la matrice de performance peut être 
        définie comme suit :
        
        >>> performance_matrix à n classes
        P(S(X_0)>S(not X_0)|H_0) P(S(X_0)>S(X_1)|H_0)     P(S(X_0)>S(X_2)|H_0)     ...
        P(S(X_1)>S(X_0)|H_1)     P(S(X_1)>S(not X_1)|H_1) P(S(X_1)>S(X_2)|H_1)     ...
        P(S(X_2)>S(X_0)|H_2)     P(S(X_2)>S(X_1)|H_2)     P(S(X_2)>S(not X_2)|H_2) ...
        ...
        
        Cas particulier, complémentarité des classes d'une classification 
        binaire (si on est pas dans l'hypothèse H_0 c'est qu'on est 
        obligatoirement dans l'hypothèse H_1) :
        
        >>> performance_matrix d'une classe binaire
        P(S(X_0)>S(X_1)|H_0)     P(S(X_0)>S(X_1)|H_0)
        P(S(X_1)>S(X_0)|H_1)     P(S(X_1)>S(X_0)|H_1)
        
        avec :
        
        >>> 
        P(S(X_0)>S(X_1)|H_0) = P(S(X_1)>S(X_0)|H_1)
        P(S(X_1)>S(X_0)|H_1) = P(S(X_0)>S(X_1)|H_0)
        P(S(X_0)>S(X_1)|H_0) + P(S(X_0)>S(X_1)|H_1) = 1
        P(S(X_0)>S(X_1)|H_0) + P(S(X_1)>S(X_0)|H_0) = 1
        
        Donc dans le cas binaire la matrice complète ne sert à rien car on 
        peut déduire l'ensemble des performances qu'avec un élément de la 
        matrice. Habituellement c'est P(S(X_1)>S(X_0)|H_1) qui est donné. 
        Il est important de noter que ceci n'est plus vrai pour une 
        classification supérieure à 2 classes car on ne peut plus exploiter 
        la complémentarité des hypothèses.
        
        See also
        --------
        process_roc, process_roc_theo
        """
        prod = list(product(range(self.nb_class), repeat=2))
        prod = chunker(prod, size)
        len_prod = int(np.ceil(self.nb_class**2/size))
        if verbose:
            prod = tqdm(
                prod,
                desc = f'{self.process_performance_matrix.__name__}',
                total = len_prod
            )
        with Parallel(n_jobs = n_jobs) as par:
            res = par(
                delayed(self._auc_par)(lot)
                for lot in prod
            )
        self.performance_matrix = pd.DataFrame(
            data = np.array(
                list(chain.from_iterable(res))
            ).reshape(self.nb_class, self.nb_class),
            columns = self.labels,
            index = self.labels
        )
        return self.performance_matrix

    def get_performance_matrix(self, **kargs):
        r"""
        Méthode de renvoit de l'attribut 'performance_matrix' si ce dernier 
        a été calculé.
        
        Parameters
        ----------
        **kargs :
            arguments de process_performance_matrix.
        
        Returns
        -------
        performance_matrix : pd.DataFrame
            DataFrame contenant la matrice de performance où 
            les noms des colonnes et indices font référence aux noms des 
            classes.
        
        See also
        --------
        process_performance_matrix
        """
        if not hasattr(self, 'performance_matrix'):
            return self.process_performance_matrix(**kargs)
        return self.performance_matrix

    def get_annalyse(self):
        r"""
        Méthode de renvoit de l'attribut 'analyse'.
        
        Returns
        -------
        analyse : Dict[str, Union[float, str]]
            pour lire facilement le contenu utilisé Pandas :
                pd.DataFrame(analyse)
        
        Notes
        -----
        Le contenu de 'analyse' varie en fonction des méthodes qui ont été 
        calculées.
        
        See also
        --------
        process_roc, process_roc_theo
        """
        return self.analyse

    def get_data_roc(self):
        r"""
        Méthode de récupération des calculs des courbes ROC. Si non calculés 
        exécute la méthode process_roc.
        
        Returns
        -------
        data_roc : List[List[float]]
            liste de calculs de chaque courbe ROC (si classification binaire 
            alors qu'une seule courbe ROC). Le format est :
                [[fpr, tpr, thresh, best_thresh, best_fpr, best_tpr, roc_auc, 
                nb_obs, idx_sel]]
        
        See also
        --------
        get_data_roc_theo, get_data_roc_bootstrap, process_roc
        """
        if not hasattr(self, 'data_roc'):
            self.process_roc()
        return self.data_roc

    def get_data_roc_bootstrap(self):
        r"""
        Méthode de récupération des calculs des courbes ROC Bootstrap. Si 
        non calculés exécute la méthode process_roc_bootstrap avec ces 
        valeurs par défaut.
        
        Returns
        -------
        data_roc_bootstrap: List[List[float]]
            liste de calculs de chaque courbe ROC par Monte-Calo. Le format 
            est :
                [[fpr, tpr, thresh, best_thresh, best_fpr, best_tpr, roc_auc, 
                nb_obs, idx_sel]]
        
        Notes
        -----
        Afin de récupérer tous les Monte-carlo par classe il suffit de faire :
        >>> data_roc_bootstrap = list(zip(*analyser.get_data_roc_bootstrap()))
        
        See also
        --------
        get_data_roc_theo, get_data_roc, process_roc_bootstrap
        """
        if not hasattr(self, 'data_roc_bootstrap'):
            self.process_roc_bootstrap()
        return self.data_roc_bootstrap

    def get_data_roc_theo(self):
        r"""
        Méthode de récupération des calculs des courbes ROC théoriques. Si non 
        calculés exécute la méthode process_roc_theo.
        
        Returns
        -------
        data_roc_theo : List[List[float]]
            liste de calculs de chaque courbe ROC (si classification binaire 
            alors qu'une seule courbe ROC). Le format est :
                [[fpr_theo, tpr_theo, thresh_theo, best_thresh_theo, 
                best_fpr_theo, best_tpr_theo, roc_auc_theo]]
        
        See also
        --------
        get_data_roc, get_data_roc_bootstrap, process_roc_theo
        """
        if not hasattr(self, 'data_roc_theo'):
            self.process_roc_theo()
        return self.data_roc_theo

    def _trim_axis(self, axis, length):
        r"""
        Méthode interne permettant de flatten les figures ainsi que de 
        supprimer les figures inutiles.
        """
        axs = axis.flat
        for ax in axs[length:]:
            ax.remove()
        return axs[:length]

    def _legend_manager(self):
        r"""
        Méthode interne permettant de rafraichir les légendes des figures.
        """
        if self.nb_class == 2:
            self.axis_roc[0].legend(loc = 4)
        else:
            for ii in range(self.nb_class):
                self.axis_roc[ii].legend(loc = 4, prop={'size': 8})

    def get_plot_roc(self):
        r"""
        Méthode de récupération des figures de la classe pour une 
        manipulation externe.
        
        Returns
        -------
        fig_roc : figure
        axis_roc : axes
        """
        return self.fig_roc, self.axis_roc

    def plot_roc_manager(self, number_columns: int=3):
        r"""
        Méthode d'initialisation des figures.
        
        Parameters
        ----------
        number_columns : int
            dans le cas d'un multi-plot permet de déterminer le nombre de 
            colonnes souhaité (le nombre de lignes sera calculé en fonction).
        
        Notes
        -----
        L'éxécution de cette méthode fait un RAZ des figures exitantes dans 
        la classe.
        """
        if number_columns > 0:
            if self.nb_class == 2:
                self.fig_roc, axis = plt.subplots(1, 1)
                self.axis_roc = np.array([axis])
                
            else:
                self.fig_roc, self.axis_roc = plt.subplots(
                    int(np.ceil(self.nb_class/number_columns)),
                    number_columns
                )
            if self.nb_class > 2:
                self.axis_roc = self._trim_axis(self.axis_roc, self.nb_class)
            for ii in range(self.nb_class) if self.nb_class > 2 else [0]:
                self.axis_roc[ii].set_aspect('equal', 'box')
                self.axis_roc[ii].grid()
        self.fig_roc.text(0.5, 0.04, 'Ratio faux positif', ha='center')
        self.fig_roc.text(
            0.04,
            0.5,
            'Ratio vrai positif',
            va='center',
            rotation='vertical'
        )
        return self

    def plot_roc(self, seuil_opti: bool=True, seuil_opti_values: bool=True):
        r"""
        Méthode d'affichage des courbes ROC dans un contexte bi-classe et 
        multi-classe.
        
        Parameters
        ----------
        seuil_opti : bool
            permet d'afficher le seuil optimal de la classe au sens de la 
            courbe ROC.
        seuil_opti_values : bool
            affiche sur la courbe la valeur du ratio de vrai positif et faux 
            positif pour le seuil optimal (utilisé seulement si 
            seuil_opti = True).
        
        Notes
        -----
        Si la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.process_roc` 
        est non éxécutée, exécution de cette dernière, idem pour la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.plot_roc_manager`.
        
        See also
        --------
        plot_roc_theo, plot_roc_bootstrap
        """
        if not hasattr(self, 'data_roc'):
            data_roc = self.process_roc().get_data_roc()
        else:
            data_roc = self.data_roc
        if not hasattr(self, 'axis_roc'):
            self.plot_roc_manager()
        for ii, mesure in enumerate(data_roc):
            if self.nb_class == 2:
                if not hasattr(self, 'data_roc_bootstrap'):
                    label = (
                        f'$P(X_1 > X_0)$ = {mesure[6]:.2}'
                        f'\nGini = {2*mesure[6]-1:.2}'
                    ) if self.labels is None else (
                        f'P($X_{self.labels[1]} > X_{self.labels[0]}$) = '
                        f'{mesure[6]:.2}'
                        f"\nGini = {2*mesure[6]-1:.2}"
                    )
                else:
                    label = (
                        f'$P(X_1>X_0)$={mesure[6]:.2f}$\pm$'
                        f"{self.analyse['roc_auc_bootstrap_std'][0]:.2f}"
                        f'\nGini={2*mesure[6]-1:.2f}$\pm$'
                        f"{2*self.analyse['roc_auc_bootstrap_std'][0]:.2f}"
                    ) if self.labels is None else (
                        f'P($X_{self.labels[1]}>X_{self.labels[0]}$)='
                        f'{mesure[6]:.2f}$\pm$'
                        f"{self.analyse['roc_auc_bootstrap_std'][0]:.2f}"
                        f'\nGini={2*mesure[6]-1:.2f}$\pm$'
                        f"{2*self.analyse['roc_auc_bootstrap_std'][0]:.2f}"
                    )
            else:
                if not hasattr(self, 'data_roc_bootstrap'):
                    label = f'AUC={mesure[6]:.2f}\nGini={2*mesure[6]-1:.2f}'
                else:
                    label = (
                        f'AUC={mesure[6]:.2f}$\pm$'
                        f"{self.analyse['roc_auc_bootstrap_std'][ii]:.2f}"
                        f'\nGini={2*mesure[6]-1:.2f}$\pm$'
                        f"{2*self.analyse['roc_auc_bootstrap_std'][ii]:.2f}"
                    )
            self.axis_roc[ii].plot(mesure[0], mesure[1], label = label)
            self.axis_roc[ii].plot([0, 1], [0, 1], 'r--', label = 'aléatoire')
            if seuil_opti:
                self.axis_roc[ii].plot(
                    [mesure[4], mesure[4]],
                    [0, mesure[5]],
                    color = '#A9A9AD',
                    linestyle = '--',
                    label = (
                        f'seuil opti {mesure[3]:.2}$\pm$'
                        f"{self.analyse['best_thresh_bootstrap_std'][ii]:.2f}"
                    ) if hasattr(self, 'data_roc_bootstrap') else (
                        f'seuil opti {mesure[3]:.2}'
                    )
                )
                self.axis_roc[ii].plot(
                    [0, mesure[4]],
                    [mesure[5], mesure[5]],
                    linestyle = '--',
                    color = '#A9A9AD'
                )
                if seuil_opti_values:
                    self.axis_roc[ii].text(
                        mesure[4],
                        0.07,
                        f"{mesure[4]:.2}",
                        rotation = -90,
                        verticalalignment = 'bottom',
                        color = '#A9A9AD',
                        fontsize = 8,
                        fontweight = 'bold'
                    )
                    self.axis_roc[ii].text(
                        0.07,
                        mesure[5],
                        f"{mesure[5]:.2}",
                        verticalalignment = 'bottom',
                        color = '#A9A9AD',
                        fontsize = 8,
                        fontweight = 'bold'
                    )
            if self.nb_class == 2:
                self.axis_roc[0].set_title(
                    "ROC curve Classe "
                    f"{'1' if self.labels is None else self.labels[1]}"
                )
            else:
                self.axis_roc[ii].set_title(
                    "ROC curve Classe "
                    f"{f'{ii}' if self.labels is None else self.labels[ii]}",
                    fontsize=10
                )
        self._legend_manager()
        return self

    def plot_roc_bootstrap(self):
        r"""
        Méthode d'affichage des courbes ROC boostrap dans un contexte 
        bi-classe et multi-classe.
        
        Notes
        -----
        Si la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.process_roc_bootstrap` 
        est non éxécutée, exécution de cette dernière, idem pour la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.plot_roc_manager`.
        
        See also
        --------
        plot_roc_theo, plot_roc
        """
        if not hasattr(self, 'data_roc_bootstrap'):
            data_roc_bootstrap = (
                self.process_roc_bootstrap().get_data_roc_bootstrap()
            )
        else:
            data_roc_bootstrap = self.data_roc_bootstrap
        if not hasattr(self, 'axis_roc'):
            self.plot_roc_manager()
        for data_roc in data_roc_bootstrap:
            for ii, mesure in enumerate(data_roc):
                self.axis_roc[ii].plot(
                    mesure[0],
                    mesure[1],
                    color = '#B5BDC4',
                    alpha = 0.5,
                    linewidth = 0.5
                )
        return self

    def plot_roc_theo(self,
                      seuil_opti: bool=True,
                      seuil_opti_values: bool=True):
        r"""
        Méthode d'affichage des estimations des courbes ROC théoriques dans un 
        contexte bi-classe et multi-classe.
        
        Parameters
        ----------
        seuil_opti : bool
            permet d'afficher le seuil optimal de la classe au sens de la 
            courbe ROC.
        seuil_opti_values : bool
            affiche sur la courbe la valeur du ratio de vrai positif et faux 
            positif pour le seuil optimal (utilisé seulement si 
            seuil_opti = True).
        
        Notes
        -----
        Si la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.process_roc_theo` 
        est non éxécutée, exécution de cette dernière, idem pour la méthode 
        :func:`~dstk.visualization._roc_analyser.ROCAnalyzer.plot_roc_manager`.
        
        See also
        --------
        plot_roc, plot_roc_bootstrap
        """
        if not hasattr(self, "data_roc_theo"):
            data_roc_theo = self.process_roc_theo().get_data_roc_theo()
        else:
            data_roc_theo = self.data_roc_theo
        if not hasattr(self, 'axis_roc'):
            self.plot_roc_manager()

        for ii, mesure in enumerate(data_roc_theo):
            if self.nb_class == 2:
                label = (
                    fr"ROC théo $P(X_1>X_0)$={mesure[6]:.2f}"
                    f"\nGini={2*mesure[6]-1:.2}"
                ) if self.labels is None else (
                    fr"ROC théo $P(X_{self.labels[1]}>X_{self.labels[0]})$="
                    f"{mesure[6]:.2f}\n"
                    f"Gini={2*mesure[6]-1:.2f}"
                )
            else:
                label = (
                    f"ROC théo AUC={mesure[6]:.2f}\n"
                    f"Gini={2*mesure[6]-1:.2f}"
                )
            self.axis_roc[ii].plot(
                mesure[0],
                mesure[1],
                label = label,
                linestyle = '--',
                color = '#83BF56'
            )
            if seuil_opti:
                self.axis_roc[ii].plot(
                    [mesure[4], mesure[4]],
                    [0, mesure[5]],
                    color = '#C9DABC',
                    linestyle = '--',
                    label = f'seuil opti théo {mesure[3]:.2}'
                )
                self.axis_roc[ii].plot(
                    [0, mesure[4]],
                    [mesure[5], mesure[5]],
                    linestyle = '--',
                    color = '#C9DABC'
                )
                if seuil_opti_values:
                    self.axis_roc[ii].text(
                        mesure[4],
                        0,
                        f"{mesure[4]:.2}",
                        rotation = -90,
                        verticalalignment = 'bottom',
                        color = '#C9DABC',
                        fontsize = 8,
                        fontweight = 'bold'
                    )
                    self.axis_roc[ii].text(
                        0,
                        mesure[5],
                        f"{mesure[5]:.2}",
                        verticalalignment = 'bottom',
                        color = '#C9DABC',
                        fontsize = 8,
                        fontweight = 'bold'
                    )
        self._legend_manager()
        return self

    def plot_roc_show(self):
        r"""
        Méthode permettant d'afficher les figures (pas besoin de relancer 
        tous les calculs si la fenêtre a été fermée).
        
        See also
        --------
        plot_roc, plot_roc_theo
        """
        self.fig_roc.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Surcharge de certaines classes de transformations Scikit-Learn pour 
s'adapter au mieux aux DataFrame.

Created on Mon Nov 23 12:19:02 2020

@author: Cyrile Delestre
"""

from warnings import warn
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.preprocessing import (OneHotEncoder, KBinsDiscretizer,
                                   RobustScaler, QuantileTransformer)

from dstk.utils import check_dataframe

class OneHotEncoderPandas(OneHotEncoder):
    r"""
    One Hot-Encoder Scikit-Learn appliqué à une DataFrame.
    Pour plus détaille voir la doc :
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html
    
    Parameters
    ----------
    subset : list
        liste de colonnes à transformer
    drop_columns : bool
        Si on préserve ou supprimer les colonnes originales
    sep : str
        séparateur entre le nom de la colonne et la classe
    
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dstk.features import OneHotEncoderPandas
    >>> 
    >>> test = pd.DataFrame(
    >>>     np.random.choice(['a', 'b', 'c', 'd'], size=10),
    >>>     columns=['tirage']
    >>> )
    >>> ohe = OneHotEncoderPandas(subset=['tirage'], drop_columns=False)
    >>> ohe.fit_transform(test)
      tirage  tirage__a  tirage__b  tirage__c  tirage__d
    0      b        0.0        1.0        0.0        0.0
    1      b        0.0        1.0        0.0        0.0
    2      d        0.0        0.0        0.0        1.0
    3      d        0.0        0.0        0.0        1.0
    4      c        0.0        0.0        1.0        0.0
    5      d        0.0        0.0        0.0        1.0
    6      c        0.0        0.0        1.0        0.0
    7      b        0.0        1.0        0.0        0.0
    8      a        1.0        0.0        0.0        0.0
    9      d        0.0        0.0        0.0        1.0
    
    Notes
    -----
    Les entrées du OneHotEncoder sont suseptible de dépendre de la version de 
    Scikit-Learn.
    """
    def __init__(self,
                 subset: list,
                 drop_columns: bool=True,
                 sep: str='__',
                 categories: str='auto',
                 drop: Optional[str]=None,
                 sparse: bool=False,
                 dtype: type=np.float64,
                 handle_unknown: str='error'):
        if not isinstance(subset, list):
            raise AttributeError(
                f"subset doit être une liste et non de type : {type(subset)}."
            )
        self.subset = subset
        self.drop_columns = drop_columns
        self.sep = sep
        if sparse:
            warn(
                "La version Pandas de OneHotEncoder ne peut pas fonctionner "
                "avec les matrices sparce. L'arguement sparse True est "
                "emplacé par False."
            )
            sparse = False
        super().__init__(
            categories=categories,
            drop=drop,
            sparse=sparse,
            dtype=dtype,
            handle_unknown=handle_unknown
        )

    def fit(self, X: pd.DataFrame, y: None=None):
        r"""\
        Méthode générique scikit-learn pour l'apprentissage.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        y : None
            pour être ISO avec Scikit-Learn
        """
        check_dataframe(X, columns_name=self.subset);
        super().fit(X[self.subset].values, y=None)
        return self

    def transform(self, X: pd.DataFrame):
        r"""\
        Méthode générique scikit-learn pour la transformation.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        """
        X = check_dataframe(X, columns_name=self.subset, copy=True)
        transfo = pd.DataFrame(
            super().transform(X[self.subset]),
            columns=[
                f"{self.subset[ii]}{self.sep}{cc}"
                    for ii, cat in enumerate(self.categories_)
                        for cc in cat
            ]
        )
        if self.drop_columns:
            X = X.drop(columns=self.subset)
        return pd.concat(
            [X.reset_index(drop=True), transfo.reset_index(drop=True)],
            axis=1
        )

class KBinsDiscretizerPandas(KBinsDiscretizer):
    r"""
    Discrétiseur KBinsDiscretizer appliqué à une DataFrame.
    Pour plus détaille voir la doc :
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KBinsDiscretizer.html#sklearn.preprocessing.KBinsDiscretizer

    Parameters
    ----------
    subset : list
        liste de colonnes à transformer
    drop : bool
        Supression ou conservation des colonnes originales (seulement pour 
        encode onehot ou onehot-dense).
    sep : str
        séparateur entre le nom de la colonne et la classe (seulement pour 
        encode onehot ou onehot-dense).
    
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dstk.features import KBinsDiscretizerPandas
    >>> 
    >>> test = pd.DataFrame(
    >>>     np.random.rand(10, 2)*99,
    >>>     columns=['age_1', 'age_2']
    >>> )
    >>> kdb = KBinsDiscretizerPandas(
    >>>     subset=['age_1', 'age_2'],
    >>>     drop=False,
    >>>     encode='onehot-dense',
    >>>     n_bins=2,
    >>>     strategy='kmeans'
    >>> )
    >>> kbd.fit_transform(test)
           age_1      age_2  age_1__0  age_1__1  age_2__0  age_2__1
    0   4.699553  14.819436       1.0       0.0       1.0       0.0
    1  81.369349  75.722283       0.0       1.0       0.0       1.0
    2  78.331536  62.132574       0.0       1.0       0.0       1.0
    3  70.266323  20.734727       0.0       1.0       1.0       0.0
    4  97.121698  66.319727       0.0       1.0       0.0       1.0
    5  56.710455  23.110782       0.0       1.0       1.0       0.0
    6  15.223127  77.464874       1.0       0.0       0.0       1.0
    7  62.688611  22.735150       0.0       1.0       1.0       0.0
    8  10.280744  56.422985       1.0       0.0       0.0       1.0
    9  58.044423  78.144402       0.0       1.0       0.0       1.0
    
    Notes
    -----
    Les entrées du OneHotEncoder sont suseptible de dépendre de la version de 
    Scikit-Learn.
    """
    def __init__(self,
                 subset: list,
                 drop: bool=True,
                 sep: str='__',
                 n_bins: int=5,
                 encode: str='onehot-dense',
                 strategy: str='quantile'):
        if not isinstance(subset, list):
            raise AttributeError(
                f"subset doit être une liste et non de type : {type(subset)}."
            )
        self.subset = subset
        self.drop = drop
        self.sep = sep
        if encode == 'onehot':
            warn(
                "La version Pandas de KNinsDiscretizer ne peut pas "
                "fonctionner avec les matrices sparce. L'arguement encode "
                "'onehot' est remplacé par 'onehot-dense'."
            )
            encode = 'onehot-dense'
        super().__init__(
            n_bins=n_bins,
            encode=encode,
            strategy=strategy
        )

    def fit(self, X: pd.DataFrame, y: None=None):
        r"""
        Méthode générique scikit-learn pour l'apprentissage.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        y : None
            pour être ISO avec Scikit-Learn
        """
        check_dataframe(X, columns_name=self.subset);
        super().fit(X[self.subset].values, y=None)
        return self

    def transform(self, X: pd.DataFrame):
        r"""
        Méthode générique scikit-learn pour la transformation.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        """
        X = check_dataframe(X, columns_name=self.subset, copy=True)
        if self.encode == 'ordinal':
            X[self.subset] = super().transform(X[self.subset])
        else:
            transfo = pd.DataFrame(
            super().transform(X[self.subset]),
            columns=[
                    f"{self.subset[ii]}{self.sep}{cc}"
                        for ii, cat in enumerate(self.n_bins_)
                            for cc in range(cat)
                ]
            )
            if self.drop:
                X = X.drop(columns=self.subset)
            X = pd.concat(
                [X.reset_index(drop=True), transfo.reset_index(drop=True)],
                axis=1
            )
        return X

class RobustScalerPandas(RobustScaler):
    r"""
    RobustScaler appliqué à une DataFrame.
    Pour plus détaille voir la doc :
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html

    Parameters
    ----------
    subset : list
        liste des colonnes à transformer
    
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dstk.features import RobustScalerPandas
    >>> 
    >>> test = pd.DataFrame(
    >>>     np.random.randn(10,3)*99+50,
    >>>     columns=['val_1', 'val_2', 'val_3']
    >>> )
    >>> RobustScalerPandas(subset=['val_1', 'val_3']).fit_transform(test)
          val_1       val_2     val_3
    0  1.928098  -86.849620 -0.809872
    1  0.655437   75.359294 -0.069656
    2 -0.734671  128.554065  0.069656
    3  0.316941  -72.130564 -0.619062
    4 -0.360382  -91.138696 -1.058226
    5  2.583596    5.003472  0.339982
    6  0.611535  134.733492  1.357780
    7 -0.837148  156.824472  0.150562
    8 -0.316941  231.724600  0.266921
    9 -0.341011  -37.310851 -1.107859
    
    Notes
    -----
    Les entrées du OneHotEncoder sont suseptible de dépendre de la version de 
    Scikit-Learn.
    """
    def __init__(self,
                 subset: list,
                 with_centering: bool=True,
                 with_scaling: bool=True,
                 quantile_range: tuple=(25.0, 75.0),
                 copy: bool=False):
        self.subset = subset
        super().__init__(
            with_centering=with_centering,
            with_scaling=with_scaling,
            quantile_range=quantile_range,
            copy=copy
        )

    def fit(self, X: pd.DataFrame, y: None=None):
        r"""
        Méthode générique scikit-learn pour l'apprentissage.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        y : None
            pour être ISO avec Scikit-Learn
        """
        check_dataframe(X, columns_name=self.subset);
        super().fit(X[self.subset].values, y=None)
        return self

    def transform(self, X: pd.DataFrame):
        r"""
        Méthode générique scikit-learn pour la transformation.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        """
        X = check_dataframe(X, columns_name=self.subset, copy=True)
        X[self.subset] = super().transform(X[self.subset])
        return X

class QuantileTransformerPandas(QuantileTransformer):
    r"""
    QuantileTransformer appliqué à une DataFrame.
    Pour plus de détaille voir la doc :
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.QuantileTransformer.html#sklearn.preprocessing.QuantileTransformer

    Parameters
    ----------
    subset : list
        liste des colonnes à transformer
    
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dstk.features import RobustScalerPandas
    >>> 
    >>> test = pd.DataFrame(
    >>>     np.random.randn(10,3)*99+50,
    >>>     columns=['val_1', 'val_2', 'val_3']
    >>> )
    >>> QuantileTransformerPandas(
    >>>     subset=['val_1', 'val_3'],
    >>>     n_quantiles=10
    >>> ).fit_transform(test)
          val_1       val_2     val_3
    0  0.888889  -86.849620  0.222222
    1  0.777778   75.359294  0.444444
    2  0.111111  128.554065  0.555556
    3  0.555556  -72.130564  0.333333
    4  0.222222  -91.138696  0.111111
    5  1.000000    5.003472  0.888889
    6  0.666667  134.733492  1.000000
    7  0.000000  156.824472  0.666667
    8  0.444444  231.724600  0.777778
    9  0.333333  -37.310851  0.000000
    
    Notes
    -----
    Les entrées du OneHotEncoder sont suseptible de dépendre de la version de 
    Scikit-Learn.
    """
    def __init__(self,
                 subset: list,
                 n_quantiles: int=1000,
                 output_distribution: str='uniform',
                 ignore_implicit_zeros: bool=False,
                 subsample: int=100000,
                 random_state: Optional[int]=None,
                 copy: bool=False):
        self.subset = subset
        super().__init__(
            n_quantiles=n_quantiles,
            output_distribution=output_distribution,
            ignore_implicit_zeros=ignore_implicit_zeros,
            subsample=subsample,
            random_state=random_state,
            copy=copy
        )

    def fit(self, X: pd.DataFrame, y: None=None):
        r"""
        Méthode générique scikit-learn pour l'apprentissage.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        y : None
            pour être ISO avec Scikit-Learn
        """
        check_dataframe(X, columns_name=self.subset);
        super().fit(X[self.subset].values, y=None)
        return self

    def transform(self, X: pd.DataFrame):
        r"""
        Méthode générique scikit-learn pour la transformation.

        Parameters
        ----------
        X : pd.DataFrame
            Observations
        """
        X = check_dataframe(X, columns_name=self.subset, copy=True)
        X[self.subset] = super().transform(X[self.subset])
        return X



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes de supervision ou de visualisation utilisant TensorBoard

Created on Sun Mar 22 11:59:21 2020

@author: Cyrile Delestre
"""

try:
    from torch.utils.tensorboard import SummaryWriter
except ModuleNotFoundError:
    raise ModuleNotFoundError(
            "Le module TensorBoard n'est pas installé : l'instruction "
            "torch.utils.tensorboard renvoit une erreur ModuleNotFoundError. "
            "Visitez la page du projet PyTorch pour savoir comment "
            "installer le package TorchVision qui comprend TensorBoard à "
            "destination de PyTorch. Ou installez directement TensorBoard "
            "via pip ou conda."
        )

from dataclasses import dataclass, field
from typing import Dict, Optional, Union, List, Tuple, Any

from numpy import ndarray, array, vectorize
from pandas import DataFrame
from torch import Tensor, zeros, from_numpy

from dstk.utils import shape_list

@dataclass
class Writer:
    r"""
    Classe d'instensiation TensorBoard qui permet d'écrire des données dans 
    TensorBoard en créant un logging TensorBoard.
    
    Parameters
    ----------
    :log_dir: Path directory où les fichiers de communication du programme 
        vers TensorBoard seront storé. Par défaut 
        chemin_courant/RUN/current_datetime_hostname qui change avec chaque 
        run.
    :comment: Ajoute un suffix au log_dir. Si log_dir est non None cet 
        argument ne sera pas pris en compte.
    :purge_step: Si le logging crache à l'étape T+X et redémarre à l'étape 
        T, tout événement qui ont un gloabl_step plus grand ou égual à X 
        seront purgé et caché dans TensorBoard.
    :max_queue: Taille de la queue d'évenement à stocker avec de pousser (add)
        vers le disque dur.
    :flush_secs: Temps de refraichissement ver les logs sur le disque dure.
    :filename_suffix: Suffix ajouter à tous les événements dans le répertoire
        log_dir.
    """
    log_dir: Optional[str]=None
    comment: str=''
    purge_step: Optional[int]=None
    max_queue: int=10
    flush_secs: int=120
    filename_suffix: str=''
    
    def get_writer(self):
        r"""
        Renvoit un logging TensorBoard.
        """
        return SummaryWriter(
            log_dir = self.log_dir,
            comment = self.comment,
            purge_step = self.purge_step,
            max_queue = self.max_queue,
            flush_secs = self.flush_secs,
            filename_suffix = self.filename_suffix
        )

@dataclass
class EmbeddingProjector:
    r"""
    Classe permettant de calculer dans TensorBoard en temps réel des 
    projecteurs dans des espaces embedded d'observations.
    
    Parameters
    ----------
    X : Union[ndarray, DataFrame, Tensor, List, Tuple]
        array like contenant les observations.
    y : Optional[Union[List,List[List], ndarray]]
        array like des labels associé à chaques observations.
    images : Optional[Union[Tensor, ndarray, List[Union[Tensor, ndarray]]]]
        images peut être de plusieur type :
            - Numpy Array ou Tensor de dimensions (N, C, H, W) avec N les 
                observation, C les channels de couleurs, H la taille et W la 
                largeur ;
            - List d'array ou de Tensor de dimension (C, H, W) ;
            - List d'array ou de Tensor de dimension (H, W).
    tag : str
        nom dans TensorBoard de l'embedding.
    kargs_writer : Dict[str, Any]
        dictionnaire d'argument pour le logging TensorBoard (voir la classe 
        Writer)
    """
    X: Union[ndarray, DataFrame, Tensor, List, Tuple]
    y: Optional[Union[List,List[List], ndarray]]=None
    image: Optional[Union[Tensor, ndarray, List[Union[Tensor, ndarray]]]]=None
    tag: str='default'
    kargs_writer: Dict[str, Any]=field(default_factory=dict)

    def __post_init__(self):
        self.writer = Writer(**self.kargs_writer).get_writer()

    def set_image(self,
                  image: Optional[Union[Tensor,
                                        ndarray,
                                        List[Union[Tensor, ndarray]]]]):
        r"""
        Permet de charger un set d'image associé aux observations.
        
        Parameters
        ----------
        image : Optional[Union[Tensor, ndarray, List[Union[Tensor, ndarray]]]]
            images peut être de plusieur type :
                - Numpy Array ou Tensor de dimensions (N, C, H, W) avec N les 
                    observation, C les channels de couleurs, H la taille et W 
                    la largeur ;
                - List d'array ou de Tensor de dimension (C, H, W) ;
                - List d'array ou de Tensor de dimension (H, W).
        """
        self.image = image
        return self

    def transform_image(self):
        r"""
        Tensorforme les images au bon format à savoir Tensor PyTorch de 
        dimensions (N, C, H, W).
        """
        image = None
        if self.image is not None:
            if isinstance(self.image, list):
                if isinstance(self.image[0], ndarray):
                    if len(self.image[0].shape) == 2:
                        dtype = from_numpy(self.image[0]).dtype
                        image = zeros(
                            (len(self.image), 1, *self.image[0].shape),
                            dtype = dtype
                        )
                        for ii, img in enumerate(self.image):
                            image[ii, 0, :, :] = from_numpy(img)
                    elif len(self.image[0].shape) == 3:
                        dtype = from_numpy(self.image[0]).dtype
                        image = zeros(
                            (len(self.image), *self.image[0].shape),
                            dtype = dtype
                        )
                        for ii, img in enumerate(self.image):
                            image[ii, :, :, :] = from_numpy(img)
                    else:
                        raise AttributeError(
                            "La dimension de l'image n'est pas comprise, "
                            "elle devrait soit être de 2 pour (H, W) ou de "
                            "3 pour (C, H, W). Mais elle est de "
                            f"{len(self.image[0].shape)}."
                        )
                elif isinstance(self.image[0], Tensor):
                    if len(self.image[0].shape) == 2:
                        image = zeros(
                            (len(self.image), 1, *self.image[0].shape),
                            dtype = self.image[0].dtype
                        )
                        for ii, img in enumerate(self.image):
                            image[ii, 0, :, :] = img
                    elif len(self.image[0].shape) == 3:
                        image = zeros(
                            (len(self.image), *self.image[0].shape),
                            dtype = self.image[0].dtype
                        )
                        for ii, img in enumerate(self.image):
                            image[ii, :, :, :] = img
                    else:
                        raise AttributeError(
                            "La dimension de l'image n'est pas comprise, "
                            "elle devrait soit être de 2 pour (H, W) ou de "
                            "3 pour (C, H, W). Mais elle est de "
                            f"{len(self.image[0].shape)}."
                        )
            if isinstance(self.image, ndarray):
                if len(self.image.shape) == 4:
                    image = from_numpy(self.image)
                else:
                    raise AttributeError(
                        "La dimension de la variable image n'est pas bonne, "
                        f"elle est de {len(self.image.shape)} alors qu'elle "
                        "devrait être de 4 avec (N, C, H, W)."
                    )
            if isinstance(self.image, Tensor):
                if len(self.image.shape) != 4:
                    raise AttributeError(
                        "La dimension de la variable image n'est pas bonne, "
                        f"elle est de {len(self.image.shape)} alors qu'elle "
                        "devrait être de 4 avec (N, C, H, W)."
                    )
        self.image = image
        return self

    def transform_X(self):
        r"""
        Transforme au format adéquate les observations.
        """
        if isinstance(self.X, DataFrame):
            self.X = self.X.values
        elif isinstance(self.X, list) and isinstance(self.X[0], list):
            self.X = array(self.X)
        elif not (isinstance(self.X, ndarray) or isinstance(self.X, Tensor)):
            raise AttributeError(
                f"L'attribue X est de type {type(self.X)}, alors qu'il "
                "devrait être soit du style Numpy Array, DataFrame ou "
                "liste de liste."
            )
        return self

    def transform_y(self):
        r"""
        Transforme en list de string les labels.
        """
        numpy_str = vectorize(
            pyfunc = lambda x: str(x),
            doc='Convert to str.',
            otypes=[str]
        )
        if isinstance(self.y, ndarray):
            self.y = numpy_str(self.y).ravel().tolist()
        elif isinstance(self.y, list):
            len_shape = len(shape_list(self.y))
            if len_shape == 2:
                self.y = [str(y[0]) for y in self.y]
            elif len_shape == 1:
                self.y = [str(y) for y in self.y]
            else:
                raise AttributeError(
                    "Si y est une liste elle doit être de dimension 1 ou 2 "
                    f"et non de dimension {len_shape}."
                )
        else:
            raise AttributeError(
                "y doit être de type Numpy Array ou liste de liste ou liste."
            )
        return self

    def push(self, global_step: Optional[int]=None, tag: Optional[str]=None):
        r"""
        Méthode permettant de pousser le projecteur dans TensorBard.
        
        Parameters
        ----------
        global_step : Optional[int]
            itération globale enregistrée
        tag : Optional[str]
            nom de l'embedding
        """
        if tag is None:
            tag = self.tag
        self.transform_X().transform_y().transform_image()
        self.writer.add_embedding(
            mat=self.X,
            metadata=self.y,
            label_img=self.image,
            global_step=global_step,
            tag=tag
        )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Fonctions de statistiques
^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Tue Nov 24 12:32:10 2020

@author: Cyrile Delestre
"""

from typing import Optional

import numpy as np

def weighted_avg_and_std(x: np.ndarray,
                         sample_weight: Optional[np.ndarray]=None,
                         axis: Optional[int]=None):
    """
    Retourne la moyenne de l'écart-type avec un poids sur chaque mesure.
    
    Parameters
    ----------
    x : np.ndarray
        variable aléatoire.
    sample_weight : Optional[np.ndarray] (=None)
        poids sur les samples. Doit avoir la même dimension que x un vecteur
        colonne.
    axis : Optional[int] (=None)
        dimension de l'opération. Si None f^ait sur toutes les dimensions.

    Returns
    -------
    mean : Union[np.ndarray, float]
        moyenne pondérée, scalaire si axis = None, sinon 
        (n1 x ... x ni x ... nq) si axis = i
    std : Union[np.ndarray, float]
        écart-type pondéré, scalaire si axis = None, sinon 
        (n1 x ... x ni x ... nq) si axis = i

    Notes
    -----
    Cette fonction retourne également la moyenne car elle en as besoin pour 
    le calcul de la variance. Il s'agit également de la moyenne pondérée.
    """
    mean = np.average(
        a=x,
        weights=sample_weight,
        axis=axis
    )
    var = np.average(
        a=(x-mean)**2,
        weights=sample_weight,
        axis=axis
    )
    if axis:
        # Obligé de faire un reshape sinon la sortie peut ne pas préserver les 
        # dimensions.
        ind_reshape = list(x.shape)
        ind_reshape[axis] = 1
        return mean.reshape(*ind_reshape), np.sqrt(var).reshape(*ind_reshape)
    return mean, np.sqrt(var)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Outils associés à la gestion SSH

Created on Sun Nov 22 12:20:19 2020

@author: Cyrile Delestre
"""

import re
from os.path import isdir

from tqdm import tqdm

def download_file(url: str,
                  file_path: str,
                  chunk_size: int=2*1024*1024,
                  header_name: str='file-name',
                  verbose: bool=True):
    r"""
    Fonction permettant de télécharger la donnée à partir d'une URL.
    
    Parameters
    ----------
    url : str
        chemin URL de la donnée.
    file_path : str
        chemin en local où la donnée téléchargé sera sauvegarder.
    chunk_size : int (=2*1024*1024)
        tronçons de traitement de la donnée (pour les données trop 
        volumineuses).
    header_name : str (='file-name')
        si file_path est un répertoire, récupère le nom de la donnée 
        directement dans les informations URL (si ce dernier existe).
    verbose : bool (=True)
        affiche une barre de progression.
    
    Notes
    -----
    Cette fonction est a utiliser de pair avec la méthode get_url_staging de 
    la classe ClientSSH. Elle permet ainsi récupérer de la donnée en local 
    directement depuis le répertoire de la cellule Hadoop.
    
    .. todo:: Améliorer la barre de progression et le débit qui est incorrecte.
    
    See also
    --------
    ClientSSH, get_url_staging
    """
    import requests
    from requests.structures import CaseInsensitiveDict
    r = requests.get(url, stream=True)
    query = requests.utils.urlparse(url).query
    params = CaseInsensitiveDict(x.split('=') for x in query.split('&'))
    file_path = re.sub('\/$','',file_path)
    file_name = ''
    if isdir(file_path) and header_name in params:
        file_name = params[header_name]
    if r.status_code == 200:
        if verbose:
            size = None
            if 'content-length' in r.headers:
                size = int(r.headers.get('content-length'))
            elif 'content-length' in params:
                size = int(r.headers.get('content-length'))
            _iter = tqdm(
                iterable = r.iter_content(chunk_size),
                desc = 'Download',
                total = size if size else None,
                unit = 'B',
                unit_scale=True,
                unit_divisor=1024
            )
        else:
            _iter = r.iter_content(chunk_size)
        with open(f"{file_path}/{file_name}", 'wb') as file:
            for chunck in _iter:
                file.write(chunck)
    else:
        IOError(
            f"La requètes get a eu pour code retour : {r.status_code}."
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes et fonctions de traitement des dates
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 09:54:33 2020

@author: Cyrile Delestre
"""

from typing import Optional, Union
from datetime import datetime as dt
from dateutil.relativedelta import relativedelta

import pandas as pd

def date_sup(x: Union[str, dt], format: Optional[str]=None):
    r"""\
    Fonction calculant une date max du mois précédent à partir d'une date au 
    format format_date.
    
    Parameters
    ----------
    x : Union[str, datetime.datetime]
        datetime object ou str, date
    format : Optional[str]
        format de la date pour conversion si la date est un string.
    
    Returns
    -------
    Retourne le jour max du mois précédent au format datetime si x est un 
    objet datetime ou string si x est un objet string.
    
    Examples
    --------
    >>> from datetime import datetime
    >>> from dstk.utils import date_sup
    >>> 
    >>> date_sup('2020-02-15', '%Y-%m-%d')
    '2020-01-31'
    >>> 
    >>> date_sup(datetime(2020, 2, 15))
    datetime.datetime(2020, 1, 31, 0, 0)
    
    See also
    --------
    date_inf
    """
    if isinstance(x, str):
        return (
            dt.strptime(x, format).replace(day=1) 
            - relativedelta(days=1)
        ).strftime(format)
    elif isinstance(x, dt):
        return x.replace(day=1) - relativedelta(days=1)
    else:
        raise AttributeError(
            "La date x doit être soit de type datetime ou str et non de type "
            f"{type(x)}."
        )

def date_inf(x: Union[str, dt], nb_month: int, format: Optional[str]=None):
    r"""
    Fonction calculant la date du premier jour du mois d'il y a nb_month mois 
    à partir d'une date au format format_date.

    Parameters
    ----------
    x : Union[str, datetime.datetime]
        datetime object ou str, date
    nb_month : int
        nombre de mois
    format : Optional[str]
        format de la date pour conversion si la date est un string.
    
    Returns
    -------
    Retourne le jour min du (nb_month)-ième mois précédant la date au format 
    datetime si x est un objet datetime ou string si x est un objet string.
    
    Examples
    --------
    >>> from datetime import datetime
    >>> from dstk.utils import date_inf
    >>> 
    >>> date_inf('2020-02-15', 3, '%Y-%m-%d')
    '2019-11-01'
    >>> 
    >>> date_inf(datetime(2020, 2, 15), 3)
    datetime.datetime(2019, 11, 1)
    
    See also
    --------
    date_sup
    """
    if isinstance(x, str):
        return (
            dt.strptime(x, format).replace(day=1) 
            - relativedelta(months=nb_month)
        ).strftime(format)
    elif isinstance(x, dt):
        return x.replace(day=1) - relativedelta(months=nb_month)
    else:
        raise AttributeError(
            "La date x doit être soit de type datetime ou str et non de type "
            f"{type(x)}."
        )

def convert_date_dataframe(X: pd.DataFrame,
                           column_date: str,
                           format: Optional[str]=None):
    r"""
    Fonction permettant de convertir une colonne d'une DataFrame en objet 
    datetime.

    Parameters
    ----------
    X : pd.DataFrame
        DataFrame.
    columns_date : str
        noms de la colonne à convertir.
    format : Optional[str]
        format de date.
    """
    if X.shape[0] > 0 and isinstance(X[column_date].iloc[0], str):
            X[column_date] = pd.to_datetime(
                X[column_date],
                format = format
            )
    return X



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes et fonctions liées à la gestion de bases SQL.

Created on Mon Nov 23 08:42:43 2020

@author: Cyrile Delestre
"""

import sqlite3
from typing import Union

import pandas as pd
import numpy as np
from tqdm import tqdm
from joblib import Parallel, delayed

def csv2sql(path_csv_or_df: Union[str, pd.DataFrame],
            path_sql: str,
            table_name: str,
            n_jobs: int=1,
            if_exists: str='append',
            **csv_opt):
    r"""
    Fonction permettant de convertir un fichier CSV en base SQLite3.
    La méthode permet de ne pas charger le CSV en RAM si ce dernier est 
    trop gros.
    
    Parameters
    ----------
    path_csv : Union[str, pd.DataFrame]
        chemin du csv ou DataFrame
    path_sql : str
        chemin de la base sql
    table_name : str
        nom de la table sql
    n_jobs : int
        nombre de jobs (gain si gros chunck), par défaut 1
    if_exists : str
        comportement si l'indice existe déjà dans la table sql. Peut être : 
        'fail', 'replace', 'append', par défaut 'append'
    **csv_opt :
        arguments de pandas.read_csv
    
    Examples
    --------
    Transfert d'un CSV trop gros pour être mis en RAM vers une base SQLite3.
    
    >>> from dstk.data import csv2sql
    >>> 
    >>> csv2sql(
    >>>     path_csv_or_df="data/mon_gros_csv.csv",
    >>>     path_sql="data/ma_table_sql.db",
    >>>     table_name="nom_de_ma_table",
    >>>     n_jobs=4,       # Nombre de jobs en parallèle
    >>>     chunksize=1000  # taille du chunker (1000 par 1000 ligne)
    >>> );
    
    Attention, dans notre cas ce sera 4 lignes qui seront stockées en RAM : 
    n_jobs*chunksize
    """
    if isinstance(path_csv_or_df, str):
        data = pd.read_csv(path_csv_or_df, **csv_opt)
    elif isinstance(path_csv_or_df, pd.DataFrame):
        data = path_csv_or_df
    else:
        raise AttributeError(
            "path_csv_or_df doit être soit un str ou une DataFrame, mais pas "
            f"{type(path_csv_or_df)}."
        )

    if "chunksize" in csv_opt.keys() and csv_opt["chunksize"] is not None:
        if isinstance(path_csv_or_df, str):
            with open(path_csv_or_df) as file:
                total = sum(1 for line in file)
        else:
            total = data.shape[0]

        @delayed
        def inser_sql(dd:pd.DataFrame):
            with sqlite3.connect(path_sql) as conn:
                dd.to_sql(
                    table_name,
                    conn,
                    if_exists = 'append',
                    index = False
                )

        with Parallel(n_jobs=n_jobs, backend='loky') as par:
            par(
                inser_sql(dd) for dd in tqdm(
                    data,
                    desc = "csv2sql",
                    total = int(np.ceil(total/csv_opt["chunksize"])),
                    miniters = 0.5
                )
            )
    else:
        with sqlite3.connect(path_sql) as conn:
            data.to_sql(
                table_name,
                conn,
                if_exists = if_exists,
                index = False
            )

def index_sql(path_sql: str,
              table_name: str,
              list_keys: list,
              index_name: str='idx',
              unique: bool=False):
    r"""
    Fonction créant un index pour une base SQLite. Créer un index peremet 
    d'accélérer grandement les filtres sur une base, car l'index est mise 
    en RAM.
    
    Parameters
    ----------
    path_sql : str
        chemin de la base sql
    table_name : str
        nom de la table sql
    list_keys: list
        list contenant les colonnes pour la génération de l'index
    index_name : str
        nom de l'index, par défaut 'idx'
    unique : bool
        si l'index est unique ou non, par défaut false
    
    Examples
    --------
    Je possède une base base.db et je sais que je vais de manière très 
    fréquente un filtrage sur la colonne A et B du type :
    
    >>> from dstk.data import PandasSQLite
    >>> 
    >>> arg1, arg2 = '...', '...'
    >>> with PandasSQLite("data/base.db") as db:
    >>>     data = db(
    >>>         "SELECT * FROM toto "
    >>>         f"WHERE A='{arg1}' AND B='{arg2}';"
    >>>     )
    
    Cette commande ira beaucoup plus vite si au préalable en aillant fait :
    
    >>> from dstk.data import index_sql
    >>> 
    >>> index_sql(
    >>>     path_sql="data/base.db",
    >>>     table_name="toto",
    >>>     list_keys=['A', 'B']
    >>> );
    
    Notes
    -----
    Commande à n'exécuter qu'une seule fois. Si plusieurs index sont créés 
    les noms 'index_name' douvent être uniques.
    """
    with sqlite3.connect(path_sql) as conn:
        cmd = (f"CREATE {'UNIQUE INDEX' if unique else 'INDEX'} {index_name} "
               f"ON {table_name}({', '.join(list_keys)});")
        cursor = conn.cursor()
        cursor.execute(cmd)
        conn.commit()
